""" SQuAD QG evaluation (sentence/answer level) """
import json
import logging
import os
from collections import defaultdict
from os.path import join as pj
from typing import Dict, List
import torch
from .automatic_evaluation_tool.bleu.bleu import Bleu
from .automatic_evaluation_tool.meteor.meteor import Meteor
from .automatic_evaluation_tool.rouge import Rouge
from .automatic_evaluation_tool.bertscore import BERTScore
from .automatic_evaluation_tool.moverscore import MoverScore
from .automatic_evaluation_tool.qa_aligned_f1_score import QAAlignedF1Score
from .spacy_module import SpacyPipeline
from .language_model import TransformersQG
from .data import get_reference_files, get_dataset

LANG_NEED_TOKENIZATION = ['ja', 'zh']


def compute_metrics(out_file,
                    tgt_file,
                    src_file: str = None,
                    prediction_aggregation: str = 'first',
                    bleu_only: bool = False,
                    language: str = 'en',
                    skip: List = None,
                    qag_model: bool = False):
    """ compute automatic metric """
    spacy_model = SpacyPipeline(language=language) if language in LANG_NEED_TOKENIZATION else None
    pairs = []
    pairs_raw = []
    with open(tgt_file, "r") as infile:
        for n, line in enumerate(infile):
            if line.endswith('\n'):
                line = line[:-1]
            pairs_raw.append({'tokenized_question': line.strip(), 'tokenized_sentence': n})
            if spacy_model is None:
                pairs.append({'tokenized_question': line.strip(), 'tokenized_sentence': n})
            else:
                pairs.append({'tokenized_question': ' '.join(spacy_model.token(line.strip())), 'tokenized_sentence': n})

    if src_file is not None and prediction_aggregation is not None:
        # group by the source (sentence where the question are produced); this is used for grouping but not evaluation
        with open(src_file, 'r') as infile:
            for n, line in enumerate(infile):
                if line.endswith('\n'):
                    line = line[:-1]
                pairs_raw[n]['tokenized_sentence'] = line.strip().lower()
                if spacy_model is None:
                    pairs[n]['tokenized_sentence'] = line.strip().lower()
                else:
                    pairs[n]['tokenized_sentence'] = ' '.join(spacy_model.token(line.strip().lower()))

    # fix prediction's tokenization: lower-casing and detaching sp characters
    with open(out_file, 'r') as infile:
        for n, line in enumerate(infile):
            if line.endswith('\n'):
                line = line[:-1]
            pairs_raw[n]['prediction'] = line.strip()
            if spacy_model is None:
                pairs[n]['prediction'] = line.strip()
            else:
                pairs[n]['prediction'] = ' '.join(spacy_model.token(line.strip()))

    # eval
    json.encoder.FLOAT_REPR = lambda o: format(o, '.4f')

    def format_pair(_pairs):
        res = defaultdict(lambda: [])
        gts = defaultdict(lambda: [])

        for pair in _pairs:

            # key is the sentence where the model generates the question
            key = pair['tokenized_sentence']

            # one generation per sentence
            if 'prediction' in pair:
                pred = pair['prediction'].encode('utf-8')
            else:
                logging.warning('prediction not found at the evaluation')
                pred = ''.encode('utf-8')
            res[key].append(pred)

            # multiple gold question per sentence
            gts[key].append(pair['tokenized_question'].encode('utf-8'))

        res_filtered = defaultdict(lambda: [])
        for k, v in res.items():
            # answer-level evaluation
            if prediction_aggregation is None:
                assert len(v) == 1
                res_filtered[k] = v
            elif prediction_aggregation == 'first':
                # the first one
                res_filtered[k] = [v[0]]
            elif prediction_aggregation == 'last':
                # the last one
                res_filtered[k] = [v[-1]]
            elif prediction_aggregation == 'long':
                # the longest generation
                res_filtered[k] = [v[v.index(sorted(v, key=len)[-1])]]
            elif prediction_aggregation == 'short':
                # the shortest generation
                res_filtered[k] = [v[v.index(sorted(v, key=len)[0])]]
            elif prediction_aggregation == 'middle':
                # middle length generation
                res_filtered[k] = [v[v.index(sorted(v, key=len)[int(len(v)/2)])]]
            else:
                raise ValueError(f'unknown aggregation method: {prediction_aggregation}')
        return gts, res_filtered

    output = {}
    scorers = [(Bleu(4), ["Bleu_1", "Bleu_2", "Bleu_3", "Bleu_4"])]
    if not bleu_only:
        scorers_extra = [
            (Meteor(), "METEOR"),
            (Rouge(), "ROUGE_L"),
            (BERTScore(language=language), "BERTScore"),
            (MoverScore(language=language), 'MoverScore')
        ]
        if qag_model:
            scorers_extra += [
                (QAAlignedF1Score(target_metric='f1', base_metric='bertscore', language=language),
                 "QAAlignedF1Score (BERTScore)"),
                (QAAlignedF1Score(target_metric='recall', base_metric='bertscore', language=language),
                 "QAAlignedRecall (BERTScore)"),
                (QAAlignedF1Score(target_metric='precision', base_metric='bertscore', language=language),
                 "QAAlignedPrecision (BERTScore)"),
                (QAAlignedF1Score(target_metric='f1', base_metric='moverscore', language=language),
                 "QAAlignedF1Score (MoverScore)"),
                (QAAlignedF1Score(target_metric='recall', base_metric='moverscore', language=language),
                 "QAAlignedRecall (MoverScore)"),
                (QAAlignedF1Score(target_metric='precision', base_metric='moverscore', language=language),
                 "QAAlignedPrecision (MoverScore)")
            ]
        if skip is not None:
            scorers_extra = [s for s in scorers_extra if s[1] not in skip]
        scorers += scorers_extra
    gts_, res_filtered_ = format_pair(pairs)
    gts_raw, res_filtered_raw = format_pair(pairs_raw)
    for scorer, method in scorers:
        if type(method) is str and method.startswith('QAAligned'):
            score, scores = scorer.compute_score(gts_raw, res_filtered_raw)
        else:
            score, scores = scorer.compute_score(gts_, res_filtered_)
        torch.cuda.empty_cache()
        if type(method) is not list:
            score, scores, method = [score], [scores], [method]
        for sc, scs, m in zip(score, scores, method):
            logging.info(f"\t{m}: {sc}")
            output[m] = sc
    return output


def evaluate(export_dir: str = '.',
             batch_size: int = 32,
             n_beams: int = 4,
             hypothesis_file_dev: str = None,
             hypothesis_file_test: str = None,
             model: str = None,
             max_length: int = 512,
             max_length_output: int = 64,
             dataset_path: str = 'lmqg/qg_squad',
             dataset_name: str = 'default',
             input_type: str = 'paragraph_answer',
             output_type: str = 'question',
             prediction_aggregation: str = 'first',
             prediction_level: str = None,
             data_caches: Dict = None,
             bleu_only: bool = False,
             overwrite: bool = False,
             use_auth_token: bool = False,
             language: str = 'en',
             test_split: str = 'test',
             validation_split: str = 'validation'):
    """ Evaluate question-generation model """
    reference_files = get_reference_files(dataset_path, dataset_name)
    if prediction_level is None:
        valid_prediction_level = [k.split('-')[0] for k in reference_files.keys()]
        if 'sentence' in valid_prediction_level:
            prediction_level = 'sentence'
        elif 'answer' in valid_prediction_level:
            prediction_level = 'answer'
        elif "questions_answers" in valid_prediction_level:
            prediction_level = 'questions_answers'
        else:
            raise ValueError(f"unexpected error: {valid_prediction_level}")
    path_metric = pj(export_dir, f'metric.{prediction_aggregation}.{prediction_level}.{input_type}.{output_type}.{dataset_path.replace("/", "_")}.{dataset_name}.json')
    metric = {}
    if not overwrite and os.path.exists(path_metric):
        with open(path_metric, 'r') as f:
            metric = json.load(f)
            if bleu_only:
                return metric
    os.makedirs(export_dir, exist_ok=True)

    if model is not None:
        lm = TransformersQG(model,
                            max_length=max_length,
                            max_length_output=max_length_output,
                            drop_overflow_error_text=False,
                            skip_overflow_error=True,
                            language=language)
        lm.eval()

        def get_model_prediction_file(split, _input_type, _output_type):
            path = pj(
                export_dir,
                f'samples.{split}.hyp.{_input_type}.{_output_type}.{dataset_path.replace("/", "_")}.{dataset_name}.txt')
            raw_input, _ = get_dataset(
                dataset_path,
                dataset_name,
                split=split,
                input_type=_input_type,
                output_type=_output_type,
                use_auth_token=use_auth_token)
            if os.path.exists(path) and not overwrite:
                with open(path) as f_read:
                    tmp = f_read.read().split('\n')
                if len(tmp) == len(raw_input):
                    return path
                logging.warning(f'recompute {path}: {len(tmp)} != {len(raw_input)}')
            prefix_type = None
            if lm.add_prefix:
                if _output_type == 'questions_answers':
                    prefix_type = 'qag'
                elif _output_type == 'question':
                    prefix_type = 'qg'
                elif _output_type == 'answer':
                    prefix_type = 'ae'
                else:
                    raise ValueError(f"prefix type is not determined for the output_type {_output_type}")
            output = lm.generate_prediction(
                raw_input,
                batch_size=batch_size,
                num_beams=n_beams,
                prefix_type=prefix_type,
                cache_path=None if data_caches is None else data_caches[split])
            with open(path, 'w') as f_writer:
                f_writer.write('\n'.join(output))
            return path

        try:
            hypothesis_file_test = get_model_prediction_file(test_split, input_type, output_type)
        except ValueError:
            hypothesis_file_test = None
        try:
            hypothesis_file_dev = get_model_prediction_file(validation_split, input_type, output_type)
        except ValueError:
            hypothesis_file_dev = None

    assert hypothesis_file_dev is not None or hypothesis_file_test is not None,\
        f'model ({model}) or file path ({hypothesis_file_dev}, {hypothesis_file_test}) is needed'

    def get_metric(split, hypothesis_file, skip_metrics):
        assert prediction_level in ['sentence', 'paragraph', 'answer'], prediction_level
        if prediction_level == 'answer':
            src_file = None
        else:
            src_file = reference_files[f'{prediction_level}-{split}']

        __metric = compute_metrics(
            out_file=hypothesis_file,
            tgt_file=reference_files[f'{output_type}-{split}'],
            src_file=src_file,
            prediction_aggregation=prediction_aggregation,
            language=language,
            bleu_only=bleu_only,
            skip=skip_metrics,
            qag_model=output_type == 'questions_answers'
        )
        return __metric

    for _hypothesis_file, _split in zip([hypothesis_file_dev, hypothesis_file_test], [validation_split, test_split]):
        if _hypothesis_file is None:
            continue
        if _split in metric:
            _metric = get_metric(_split, _hypothesis_file, list(metric[_split].keys()))
            metric[_split].update(_metric)
        else:
            metric[_split] = get_metric(_split, _hypothesis_file, None)

    with open(path_metric, 'w') as f:
        json.dump(metric, f)
    return metric





import os
import json
import re
from os.path import join as pj
from glob import glob
import pandas as pd
from lmqg.language_model import TASK_PREFIX
from datasets import load_dataset

bib = """@inproceedings{ushio-etal-2022-generative,
    title = "{G}enerative {L}anguage {M}odels for {P}aragraph-{L}evel {Q}uestion {G}eneration",
    author = "Ushio, Asahi  and
        Alva-Manchego, Fernando  and
        Camacho-Collados, Jose",
    booktitle = "Proceedings of the 2022 Conference on Empirical Methods in Natural Language Processing",
    month = dec,
    year = "2022",
    address = "Abu Dhabi, U.A.E.",
    publisher = "Association for Computational Linguistics",
}
"""
paper_link = "[https://arxiv.org/abs/2210.03992](https://arxiv.org/abs/2210.03992)"
version_description = {
    'default': "This model is fine-tuned without parameter search (default configuration is taken from [ERNIE-GEN](https://arxiv.org/abs/2001.11314)).",
    'no-answer': "This model is fine-tuned without answer information, i.e. generate a question only given a paragraph (note that normal model is fine-tuned to generate a question given a pargraph and an associated answer in the paragraph).",
    'no-paragraph': "This model is fine-tuned without pargraph information but only the sentence that contains the answer.",
}
sample_qa_dict = {
    "en": [
        "question: What is a person called is practicing heresy?, context: Heresy is any provocative belief or theory that is strongly at variance with established beliefs or customs. A heretic is a proponent of such claims or beliefs. Heresy is distinct from both apostasy, which is the explicit renunciation of one's religion, principles or cause, and blasphemy, which is an impious utterance or action concerning God or sacred things.",
        "question: who created the post as we know it today?, context: 'So much of The Post is Ben,' Mrs. Graham said in 1994, three years after Bradlee retired as editor. 'He created it as we know it today.'— Ed O'Keefe (@edatpost) October 21, 2014"],
    "ja": [
        "question: 新型車両として6000系が構想されたのは、製造費用のほか、どんな費用を抑えるためだったの?, context: 三多摩地区開発による沿線人口の増加、相模原線延伸による多摩ニュータウン乗り入れ、都営地下鉄10号線(現都営地下鉄新宿線、以下新宿線と表記する)乗入構想により、京王線の利用客増加が見込まれ、相当数の車両を準備する必要に迫られるなか、製造費用、保守費用を抑えた新型車両として6000系が構想された。新宿線建設に際してはすでに1号線(後の浅草線)を1,435mm軌間で開業させていた東京都は京成電鉄と1号線との乗り入れにあたり京成電鉄の路線を1,372mmから1,435mmに改軌させた事例や、1,372mm軌間の特殊性から運輸省(当時、2001年から国土交通省)と共に京王にも改軌を求めたが、改軌工事中の輸送力確保が困難なことを理由に改軌しないことで決着している。",
        "question: 1968年に開催されたオリンピックの名前は何ですか?, context: オリンピックが世界的大イベントに成長するに従って政治に左右されるようになると、1968年のメキシコシティ大会では黒人差別を訴える場と化し、1972年のミュンヘン大会ではアラブのゲリラによるイスラエル選手に対するテロ事件まで起きた(ミュンヘンオリンピック事件)。1976年のモントリオール大会になると、ニュージーランドのラグビーチームの南アフリカ遠征に反対してアフリカの諸国22ヶ国がボイコットを行った。そして、1980年のモスクワ大会ではソ連のアフガニスタン侵攻に反発したアメリカ・西ドイツ・日本などの西側諸国が相次いでボイコットを行った。1984年ロサンゼルス大会ではソ連と東側諸国が報復ボイコットを行ない、参加したのはソ連と対立していた中国とルーマニアだけだった。中でも、イラン革命後のイラン・イスラム共和国はモスクワとロサンゼルス双方のオリンピックをボイコットしている。オリンピックが巨大化するに従って財政負担の増大が大きな問題となり、1976年の夏季大会では大幅な赤字を出し、その後夏季・冬季とも立候補都市が1〜2都市だけという状態が続いた。"
    ],
    "ru": [
        "question: чем соответствует абсолютная погрешность скорости света ?, context: Наивысшая точность измерений была достигнута в начале 1970-х. В 1975 году XV Генеральная конференция по мерам и весам зафиксировала это положение и рекомендовала считать скорость света, равной 299 792 458 м/с с относительной погрешностью 4•10−9, что соответствует абсолютной погрешности 1,1 м/с. Впоследствии это значение скорости света было положено в основу определения метра в Международной системе единиц (СИ), а сама скорость света стала рассматриваться как фундаментальная физическая постоянная, по определению равная указанному значению точно.",
        "question: Какие начинания предпринял Lloyds в начале 1970-х годов?, context: В начале 1970-х Lloyds начал расширять деятельность на международной арене, для чего был создан Lloyds Bank International. География его деятельности включала ФРГ, Швейцарию, Ближний Восток, Австралию, Канаду и США; к 1978 году Lloyds был представлен в 43 странах. В 1972 году было создано подразделение страхования, а в 1973 году была основана лизинговая компания Lloyds Leasing. В 1979 году банк начал предоставлять услуги ипотечного кредитования (при покупке недвижимости стоимостью от £25 000 до £150 000). В 1982 году начало работу агентство недвижимости Blackhorse Agencies, к 1989 году у него было 563 отделения. В 1986 году сфера деятельности Lloyds Bank PLC ещё больше расширилась с учреждением брокерской конторы и торгового банка Lloyds Merchant Bank. В 1988 году была поглощена страховая компания Abbey Life Group PLC; после объединения с ней всей своей страховой деятельности была образована дочерняя компания Lloyds Abbey Life. В 1995 году Lloyds Bank Plc объединился с TSB Group plc (группой, образованной в 1986 году из четырёх сберегательных банков Trustee Savings Banks) под названием Lloyds TSB Bank plc. В 2000 году за £7 млрд была поглощена шотландская взаимная страховая компания Scottish Widows."
    ],
    "ko": [
        "question: 매드 클라운이 참가해 큰 화제를 모았던 프로그램은?, context: 과거 소울 컴퍼니 소속으로 소울 컴퍼니 해체 후 현재의 소속사는 스타쉽 엑스이다. Mad Clown vs Crucial Star (매드 클라운 vs 크루셜 스타)라는 프로젝트 그룹으로 크루셜 스타와 함께 활동하기도 하였으며, 2013년부터는 MC인 저스디스와 팀을 이루어 랩 듀오 커먼콜드로 활동하고 있다. 또한 Mnet 《쇼미더머니 2》에서 참가자로 참가하여 큰 화제를 모았으며, 《쇼미더머니 5》에서는 길 & 매드 클라운 팀으로 프로듀서로 출연하였다., 재발매 물량도 완판되어 추가 제작에 들어갔다. 2016년 4월, 소속사와 자신의 SNS를 통해 2016년 5월 15일 현재 교제 중인 일반인 여자친구와의 결혼을 공식발표하였다.",
        "question: 1913년 필라델피아 애슬레틱스의 개막전 상대는?, context: 1913년 시즌을 앞두고 스프링 트레이닝에서 잭 쿰스는 앨라배마 주 몽고메리에서 고열로 힘들어했는데, 당시에는 식중독 및 늑막염 진단을 받고 휴식을 취했다. 4월 10일, 보스턴 레드삭스를 상대로 치러진 개막전에서 잭 쿰스는 선발투수로 내정되었다. 그는 3이닝을 노히트로 막고 6회 치프 벤더와 교체되었으며, 경기는 10-5로 애슬레틱스가 승리했다. 이틀 뒤에 다시 선발 등판에 나섰으나 ⁄3이닝 동안 2피안타 1볼넷, 4실점만을 기록하고 강판되었다. 쿰스는 보스턴에서의 시리즈를 끝내고 팀 동료들과 함께 워싱턴으로 향했지만, 고통이 심해지자 구단은 그를 필라델피아로 돌려보냈다. 그곳에서 그는 장티푸스 진단을 받고 휴식을 취했으며, 8월에 다시 팀에 복귀하려고 했지만 정상적인 회복을 위해서 다시 병원에 들어갔다. 이 기간 몸무게가 25 kg 가량이나 감소했다. 이 해 필라델피아 애슬레틱스는 월드 시리즈에서 2년만에 다시 뉴욕 자이언츠와 맞붙었고, 우승을 차지했다. 쿰스의 공백기는 다음해인 1914년 시즌까지 길어졌다. 이 해 시즌에는 팀 순위가 정해진 시즌 막판에야 두 경기에 선발 출전해서, 도합 8이닝 8피안타 4실점, 4.50의 평균자책점을 기록했다. 시즌 후인 12월 9일, 애슬레틱스에서 방출되었다."
    ],
    "es": [
        "question: ¿Cuál es la población de Nueva York a partir de 2014?, context: Situada en uno de los mayores puertos naturales del mundo, la ciudad de Nueva York consta de cinco municipios, cada uno de los cuales es un condado separado del estado de Nueva York. Los cinco distritos - Brooklyn, Queens, Manhattan, el Bronx y Staten Island - se consolidaron en una sola ciudad en 1898. Con una población censada estimada en 2014 de 8.491.079 habitantes distribuidos en una superficie de solo 790 km ², Nueva York es la ciudad más densamente poblada de los Estados Unidos. Hasta 800 idiomas se hablan en Nueva York, por lo que es la ciudad más lingüísticamente diversa del mundo. Según estimaciones del censo de 2014, la región metropolitana de la ciudad de Nueva York sigue siendo por un margen significativo la más poblada de los Estados Unidos, según lo definido tanto por el Área Estadística Metropolitana (20,1 millones de residentes). En 2013, el MSA produjo un producto metropolitano bruto (GMP) de casi US $1,39 billones, mientras que en 2012, el CSA generó un GMP de más de US $1,55 billones, ambos clasificados en primer lugar.",
        "question: ¿Cómo se llama el ejército personal de Sassou?, context: El progreso democrático del Congo se descarriló en 1997, cuando Lissouba y Sassou comenzaron a luchar por el poder en la guerra civil. A medida que se acercaban las elecciones presidenciales de julio de 1997, las tensiones entre los campos de Lissouba y Sassou aumentaron. El 5 de junio, las fuerzas del gobierno del presidente Lissouba rodearon el complejo de Sassou en Brazzaville y Sassou ordenó a los miembros de su milicia privada (conocida como Cobras) resistir. Así comenzó un conflicto de cuatro meses que destruyó o dañó gran parte de Brazzaville y causó decenas de miles de muertes civiles. A principios de octubre, el régimen socialista angoleño comenzó una invasión del Congo para instalar a Sassou en el poder. A mediados de octubre, el gobierno de Lissouba cayó. Poco después, Sassou se declaró presidente."
    ],
    "fr": [
        "question: En quelle année a-t-on trouvé trace d'un haut fourneau similaire?, context: Cette technologie ne disparaît qu'au début du XXe siècle. On retrouve vers 1900 un haut fourneau similaire dans le Bulacan, aux Philippines. Plus tard encore, le « haut fourneau dans la cour » prôné par Mao Zedong pendant le Grand Bond en avant est de ce type. L'expérience n'est un échec technique que dans les régions où le savoir-faire n'existe pas, ou a disparu.",
        "question: Comment appelle-t-on la Guerre de 14-18 ?, context: Ce black dog peut être lié à des évènements traumatisants issus du monde extérieur, tels que son renvoi de l'Amirauté après la catastrophe des Dardanelles, lors de la Grande Guerre de 14-18, ou son rejet par l'électorat en juillet 1945. On sait également que dans ces deux cas, la guérison, certes lente et douloureuse et jamais complète ni définitive, se fera grâce à la peinture. D'un autre côté, étant donnés les symptômes de ce mal que Churchill éprouvait de plus en plus, il ne pouvait rien moins qu'être purement associé à de telles causes extrinsèques, ce qui correspond au profil classique de la dépression majeure unipolaire ou bipolaire."

    ],
    "de": [
        "question: Welche Auszeichnung hat die Wartburg 1999 erhalten?, context: Thüringen == Kultur == Die Kulturlandschaft Thüringens ist bedingt durch die lange politische Zersplitterung (bis 1920) recht vielfältig. Diese Vielfalt hat sich bis heute erhalten und findet in den verschiedenen ehemaligen Residenzen im Land mit ihren historisch gewachsenen Museen und Theatern Ausdruck. Parallel zur Vielfalt der Landesteile verbinden aber vor allem die ähnliche Küche sowie ähnlichen Feste und Bräuche. Prägend für die Kultur sind nach wie vor die zahlreichen Stätten der klassischen Hochkultur von der Reformation bis zum Bauhaus hinter denen die Orte der Gegenwartskultur ein Stück weit zurückfallen. Zum UNESCO-Welterbe in Thüringen gehören seit 1996 die Bauhaus-Stätten in Weimar mit dem zwischen 1904 und 1911 nach Plänen von Henry van de Velde errichteten Hauptgebäude der Bauhaus-Universität, der Kunstgewerbeschule Weimar und dem Musterhaus Am Horn, seit 1998 die elf Stätten des Klassischen Weimars (Goethes Wohnhaus, Schillers Wohnhaus, Herderkirche und Herder-Stätten, Weimarer Stadtschloss, Wittumspalais, Herzogin Anna Amalia Bibliothek, Park an der Ilm mit Goethes Gartenhaus und Römischem Haus, Schloss Belvedere, Schloss Ettersburg, Schloss Tiefurt, Historischer Friedhof Weimar), seit 1999 die Wartburg bei Eisenach und seit 2011 der Nationalpark Hainich als Teil der Europäischen Buchenurwälder." ,
        "question: Wann endete die Aberdeen Regierung? , context: Krimkrieg === Großbritannien === Der Krimkrieg zeigte, dass es erhebliche Missstände im britischen Militär gab. Dadurch verlor die Regierung Aberdeen erheblich an Ansehen. Im Februar 1855 wurde sie zum Rücktritt gezwungen, und Palmerston übernahm die Bildung eines neuen Kabinetts. Der spätere britische Premierminister Disraeli erklärte den Krieg aus einer von Südasien eingenommenen Perspektive zu einem „indischen Krieg“, da es zuvor (irreale) Befürchtungen gegeben hatte, dass Russland durch eine Expansion nach Süden das britische Indien in Gefahr bringen könnte. Das Verhältnis zwischen Großbritannien und Russland blieb bis ins 20. Jahrhundert aus ideologischen und weltmachtpolitischen Gründen angespannt. Der Krieg führte in Großbritannien zur Bildung eines modernen Nationalmythos des die Ehre der Nation verteidigenden „gemeinen“ Soldaten, anstelle des Aristokraten früherer Kriege. In der Mittelklasse kam es zu einem neuen Gefühl des Selbstbewusstseins im Zusammenhang von Ideen wie professioneller Fähigkeit und dem Leistungsprinzip. Die Mittelklasse erkannte sich in einer Florence Nightingale wieder, die zur Nationalheldin aufstieg. Die Königin stiftete 1857 das Victoria-Kreuz, mit dem erstmals Nichtoffiziere ausgezeichnet werden konnten."
    ],
    "it": [
        "question: Quale batterio ha il nome del paese che colpisce di più nel suo nome?, context: Il complesso M. tubercolosi (MTBC) comprende altri quattro micobatteri causa di tubercolosi: M. bovis, M. africanum, M. canetti e M. microti. M. africanum non è molto diffuso, ma è una causa significativa di tubercolosi in alcune parti dell' Africa. M. bovis era una volta una causa comune della tubercolosi, ma l' introduzione del latte pastorizzato ha quasi completamente eliminato questo problema di salute pubblica nei paesi sviluppati. M. canetti è raro e sembra essere limitato al Corno d' Africa, anche se alcuni casi sono stati osservati negli emigranti africani. M. microti è anche raro ed è visto quasi solo in persone immunodeficienti, anche se la sua prevalenza può essere significativamente sottovalutata.",
    ]
}
sample_qg_dict = {
    "en": [
        "<hl> Beyonce <hl> further expanded her acting career, starring as blues singer Etta James in the 2008 musical biopic, Cadillac Records.",
        "Beyonce further expanded her acting career, starring as blues singer <hl> Etta James <hl> in the 2008 musical biopic, Cadillac Records.",
        "Beyonce further expanded her acting career, starring as blues singer Etta James in the 2008 musical biopic,  <hl> Cadillac Records <hl> ."
        ],
    "ja": [
        "ゾフィーは貴族出身ではあったが王族出身ではなく、ハプスブルク家の皇位継承者であるフランツ・フェルディナントとの結婚は貴賤結婚となった。皇帝フランツ・ヨーゼフは、2人の間に生まれた子孫が皇位を継がないことを条件として結婚を承認していた。視察が予定されている<hl>6月28日<hl>は2人の14回目の結婚記念日であった。",
        "『クマのプーさん』の物語はまず1925年12月24日、『イヴニング・ニュース』紙のクリスマス特集号に短編作品として掲載された。これは『クマのプーさん』の第一章にあたる作品で、このときだけは挿絵をJ.H.ダウドがつけている。その後作品10話と挿絵が整い、刊行に先駆けて「イーヨーの誕生日」のエピソードが1926年8月に『ロイヤルマガジン』に、同年10月9日に『ニューヨーク・イヴニング・ポスト』紙に掲載されたあと、同年10月14日にロンドンで(メシュエン社)、21日にニューヨークで(ダットン社)『クマのプーさん』が刊行された。前著『ぼくたちがとてもちいさかったころ』がすでに大きな成功を収めていたこともあり、イギリスでは初版は前著の7倍に当たる<hl>3万5000部<hl>が刷られた。他方のアメリカでもその年の終わりまでに15万部を売り上げている。ただし依然として人気のあった前著を売り上げで追い越すには数年の時間を要した。",
        "フェルメールの作品では、17世紀のオランダの画家、ヨハネス・フェルメールの作品について記述する。フェルメールの作品は、疑問作も含め<hl>30数点<hl>しか現存しない。現存作品はすべて油彩画で、版画、下絵、素描などは残っていない。以下には若干の疑問作も含め、37点の基本情報を記載し、各作品について略説する。収録順序、推定制作年代は『「フェルメールとその時代展」図録』による。日本語の作品タイトルについては、上掲図録のほか、『「フェルメール展」図録』、『フェルメール生涯と作品』による。便宜上「1650年代の作品」「1660年代の作品」「1670年代の作品」の3つの節を設けたが、フェルメールの作品には制作年代不明のものが多く、推定制作年代については研究者や文献によって若干の差がある。",
    ],
    "ru": [
        "Нелишним будет отметить, что, развивая это направление, Д. И. Менделеев, поначалу априорно выдвинув идею о температуре, при которой высота мениска будет нулевой, <hl> в мае 1860 года <hl> провёл серию опытов.",
        "Однако, франкоязычный <hl> Квебек <hl> практически никогда не включается в состав Латинской Америки.",
        "Классическим примером международного синдиката XX века была группа компаний <hl> Де Бирс <hl> , которая в 1980-е годы контролировала до 90 % мировой торговли алмазами."
    ],
    "ko": [
        "1990년 영화 《 <hl> 남부군 <hl> 》에서 단역으로 영화배우 첫 데뷔에 이어 같은 해 KBS 드라마 《지구인》에서 단역으로 출연하였고 이듬해 MBC 《여명의 눈동자》를 통해 단역으로 출연하였다.",
        "백신이 없기때문에 예방책은 <hl> 살충제 <hl> 를 사용하면서 서식 장소(찻찬 받침, 배수로, 고인 물의 열린 저장소, 버려진 타이어 등)의 수를 줄임으로써 매개체를 통제할 수 있다.",
        "<hl> 원테이크 촬영 <hl> 이기 때문에 한 사람이 실수를 하면 처음부터 다시 찍어야 하는 상황이 발생한다."
    ],
    "es": [
        "del <hl> Ministerio de Desarrollo Urbano <hl> , Gobierno de la India.",
        "a <hl> noviembre <hl> , que es también la estación lluviosa.",
        "como <hl> el gobierno de Abbott <hl> que asumió el cargo el 18 de septiembre de 2013."
    ],
    "fr": [
        "Créateur » (Maker), lui aussi au singulier, « <hl> le Suprême Berger <hl> » (The Great Shepherd) ; de l'autre, des réminiscences de la théologie de l'Antiquité : le tonnerre, voix de Jupiter, « Et souvent ta voix gronde en un tonnerre terrifiant », etc.",
        "Ce black dog peut être lié à des évènements traumatisants issus du monde extérieur, tels que son renvoi de l'Amirauté après la catastrophe des Dardanelles, lors de la <hl> Grande Guerre <hl> de 14-18, ou son rejet par l'électorat en juillet 1945.",
        "contre <hl> Normie Smith <hl> et 15 000 dollars le 28 novembre 1938."
    ],
    "de": [
        "Empfangs- und Sendeantenne sollen in ihrer Polarisation übereinstimmen, andernfalls <hl> wird die Signalübertragung stark gedämpft. <hl>",
        "das erste weltweit errichtete Hermann Brehmer <hl> 1855 <hl> im niederschlesischen ''Görbersdorf'' (heute Sokołowsko, Polen).",
        "Er muss Zyperngrieche sein und wird direkt für <hl> fünf Jahre <hl> gewählt (Art. 43 Abs. 1 der Verfassung) und verfügt über weitreichende Exekutivkompetenzen."
    ],
    "it": [
        "<hl> Dopo il 1971 <hl> , l' OPEC ha tardato ad adeguare i prezzi per riflettere tale deprezzamento.",
        "L' individuazione del petrolio e lo sviluppo di nuovi giacimenti richiedeva in genere <hl> da cinque a dieci anni <hl> prima di una produzione significativa.",
        "il <hl> Giappone <hl> è stato il paese più dipendente dal petrolio arabo."
    ]
}
sample_ae_dict = {
    "en": [
        "<hl> Beyonce further expanded her acting career, starring as blues singer Etta James in the 2008 musical biopic, Cadillac Records. <hl> Her performance in the film received praise from critics, and she garnered several nominations for her portrayal of James, including a Satellite Award nomination for Best Supporting Actress, and a NAACP Image Award nomination for Outstanding Supporting Actress.",
        "Beyonce further expanded her acting career, starring as blues singer Etta James in the 2008 musical biopic, Cadillac Records. <hl> Her performance in the film received praise from critics, and she garnered several nominations for her portrayal of James, including a Satellite Award nomination for Best Supporting Actress, and a NAACP Image Award nomination for Outstanding Supporting Actress. <hl>"
    ],
    "ja": [
        "『クマのプーさん』の物語はまず1925年12月24日、『イヴニング・ニュース』紙のクリスマス特集号に短編作品として掲載された。これは『クマのプーさん』の第一章にあたる作品で、このときだけは挿絵をJ.H.ダウドがつけている。その後作品10話と挿絵が整い、刊行に先駆けて「イーヨーの誕生日」のエピソードが1926年8月に『ロイヤルマガジン』に、同年10月9日に『ニューヨーク・イヴニング・ポスト』紙に掲載されたあと、同年10月14日にロンドンで(メシュエン社)、21日にニューヨークで(ダットン社)『クマのプーさん』が刊行された。<hl>前著『ぼくたちがとてもちいさかったころ』がすでに大きな成功を収めていたこともあり、イギリスでは初版は前著の7倍に当たる3万5000部が刷られた。<hl>他方のアメリカでもその年の終わりまでに15万部を売り上げている。ただし依然として人気のあった前著を売り上げで追い越すには数年の時間を要した。",
        "フェルメールの作品では、17世紀のオランダの画家、ヨハネス・フェルメールの作品について記述する。フェルメールの作品は、疑問作も含め30数点しか現存しない。<hl>現存作品はすべて油彩画で、版画、下絵、素描などは残っていない。以下には若干の疑問作も含め、37点の基本情報を記載し、各作品について略説する。<hl>収録順序、推定制作年代は『「フェルメールとその時代展」図録』による。日本語の作品タイトルについては、上掲図録のほか、『「フェルメール展」図録』、『フェルメール生涯と作品』による。便宜上「1650年代の作品」「1660年代の作品」「1670年代の作品」の3つの節を設けたが、フェルメールの作品には制作年代不明のものが多く、推定制作年代については研究者や文献によって若干の差がある。"
    ],
    "ru": [
        "<hl> в английском языке в нарицательном смысле применяется термин rapid transit (скоростной городской транспорт), однако употребляется он только тогда, когда по смыслу невозможно ограничиться названием одной конкретной системы метрополитена. <hl> в остальных случаях используются индивидуальные названия: в лондоне — london underground, в нью-йорке — new york subway, в ливерпуле — merseyrail, в вашингтоне — washington metrorail, в сан-франциско — bart и т. п. в некоторых городах применяется название метро (англ. metro) для систем, по своему характеру близких к метро, или для всего городского транспорта (собственно метро и наземный пассажирский транспорт (в том числе автобусы и трамваи)) в совокупности.",
        "Вопреки ожиданиям, объединение денежных систем республик не привело к уменьшению инфляции. Напротив, закдензнаки стали невероятно быстро обесцениваться, особенно в 1924 году. Для обеспечения денежного рынка приходилось увеличивать эмиссию закдензнаков и выпускать в оборот купюры невероятно больших номиналов. <hl> Так, в период с 1 января по 20 марта 1924 года были введены в оборот купюры достоинством 25 000 000 рублей, затем — 250 000 000 рублей. <hl> И, наконец, в апреле 1924 года были выпущены купюры миллиардного достоинства (в просторечии лимард)."
    ],
    "ko": [
        "또한 스피어스는 많은 새로운 여성 아티스트들에게 영향을 끼쳤는데, 대표적으로 데미 로바토, 케이티 페리, 크리스티니아 드바지, 레이디 가가, 리틀 부츠, 셀레나 고메즈 & 더씬, 픽시 로트 이 있다. 2007년 비욘세 놀스는 Total Request Live와의 인터뷰에서 '나는 브리트니를 사랑하고 팬이에요. 특히 새 앨범 Blackout을 좋아해요'라고 말했다. 린제이 로한은 '언제나 브리트니 스피어스에게 영감을 받는다. 학창시절 그녀처럼 타블로이드에 오르기를 꿈꿔왔다'고 말하며 롤 모델로 꼽았다. 스피어스는 현대 음악가들에게 음악적 영감으로 언급되기도 했다. <hl> 마일리 사이러스는 자신의 히트곡 Party in the U.S.A. 가 브리트니에게 영감과 영향을 받은 곡이라고 밝혔다. <hl> 베리 매닐로우의 앨범 15 Minutes 역시 브리트니에게 영감을 얻었다고 언급되었다.",
        "지난 22일 아프리카TV는 BJ 철구가 서비스 정지 처분을 받았음을 밝혔다. 서비스 정지 처분을 사유는 철구가 10대 청소년에게 유해한 장면을 방송으로 내보냈기 때문이었다. 문제가 된 장면은 BJ 철구가 미성년자는 시청할 수 없게 하는 19세 시청 가능 설정을 하지 않은 채 흡연하는 모습을 여과 없이 드러낸 장면이다. 아프리카TV는 청소년 보호 정책의 '청소년들이 해로운 환경으로부터 보호받을 수 있도록 조치한다'라고 조항을 근거로 철구에게 서비스 정지 처분을 내렸다. 흡연 이외에 음주 방송 등도 19세 시청 가능 설정을 해야만 방송할 수 있다. <hl> 게다가 철구의 방송 정지 처분은 이번에 처음이 아니라 16번 째기 때문에 더욱더 논란이 되고 있다. <hl>"
    ],
    "es": [
        "<hl> En la diáspora somalí, múltiples eventos islámicos de recaudación de fondos se llevan a cabo cada año en ciudades como Birmingham, Londres, Toronto y Minneapolis, donde los académicos y profesionales somalíes dan conferencias y responden preguntas de la audiencia. <hl> El propósito de estos eventos es recaudar dinero para nuevas escuelas o universidades en Somalia, para ayudar a los somalíes que han sufrido como consecuencia de inundaciones y / o sequías, o para reunir fondos para la creación de nuevas mezquitas como.",
        "<hl> Los estudiosos y los histori a dores están divididos en cuanto a qué evento señala el final de la era helenística. <hl> El período helenístico se puede ver que termina con la conquista final del corazón griego por Roma en 146 a. C. tras la guerra aquea, con la derrota final del reino ptolemaico en la batalla de Actium en 31 a. Helenístico se distingue de helénico en que el primero abarca toda la esfera de influencia griega antigua directa, mientras que el segundo se refiere a la propia Grecia."
    ],
    "fr": [
        "Pourtant, la strophe spensérienne, utilisée cinq fois avant que ne commence le chœur, constitue en soi un vecteur dont les répétitions structurelles, selon Ricks, relèvent du pur lyrisme tout en constituant une menace potentielle. Après les huit sages pentamètres iambiques, l'alexandrin final <hl> permet une pause <hl>, « véritable illusion d'optique » qu'accentuent les nombreuses expressions archaïsantes telles que did swoon, did seem, did go, did receive, did make, qui doublent le prétérit en un temps composé et paraissent à la fois « très précautionneuses et très peu pressées ».",
        "Néanmoins, une fois encore, l'arithmétique modulaire est insuffisante pour venir à bout du théorème. Dirichlet utilise de nombreuses techniques analytiques, comme les séries entières et l'analyse complexe. Le fruit de ces travaux donne naissance à une nouvelle branche des mathématiques : la théorie analytique des nombres. L'un des points cruciaux de cette théorie provient de l'unique article de <hl> Bernhard Riemann <hl> en théorie des nombres : Sur le nombre de nombres premiers inférieurs à une taille donnée. Il conjecture une localisation des racines de sa fonction ζ. La recherche de la position des racines, initiée par Dirichlet, devient une préoccupation centrale et reste l'une des conjectures pressenties comme les plus difficiles des mathématiques de notre époque.",
    ],
    "de": [
        "Sommerzeit <hl> Frühling <hl>: Umstellung von Normalzeit auf Sommerzeit – die Uhr wird um eine Stunde ''vor''gestellt. Herbst: Umstellung von Sommerzeit auf Normalzeit – die Uhr wird um eine Stunde ''zurück''gestellt. Als Sommerzeit wird die gegenüber der Zonenzeit meist um eine Stunde vorgestellte Uhrzeit bezeichnet, die während eines bestimmten Zeitraums im Sommerhalbjahr (und oft auch etwas darüber hinaus) als gesetzliche Zeit dient. Eine solche Regelung wird fast nur in Ländern der gemäßigten Zonen angewandt. Die mitteleuropäische Sommerzeit beginnt am letzten Sonntag im März um 2:00 Uhr MEZ, indem die Stundenzählung um eine Stunde von 2:00 Uhr auf 3:00 Uhr vorgestellt wird. Sie endet jeweils am letzten Sonntag im Oktober um 3:00 Uhr MESZ, indem die Stundenzählung um eine Stunde von 3:00 Uhr auf 2:00 Uhr zurückgestellt wird.",
        "Iran === Landwirtschaft === Die landwirtschaftliche Nutzfläche beträgt trotz zahlreicher Gebirge und Wüsten 10 % der Landesfläche, wobei ein Drittel künstlich bewässert wird. Die Landwirtschaft ist einer der größten Arbeitgeber des Landes. Wichtige Produkte sind Pistazien, Weizen, Reis, Zucker, Baumwolle, Früchte, Nüsse, Datteln, Wolle und Kaviar. Seit der Revolution von 1979 wurde der Anbau von Weintrauben wegen des islamischen Alkoholverbots auf den 200.000 Hektar Rebfläche fast vollständig auf Tafeltrauben und Rosinen umgestellt. Bei Rosinen ist <hl> der Iran <hl> inzwischen nach der Türkei der zweitgrößte Exporteur der Welt, bei Safran mit ungefähr 90 % Marktanteil des globalen Bedarfs mit Abstand der größte."
    ],
    "it": [
        "<hl> Il 6 ottobre 1973 , la Siria e l' Egitto, con il sostegno di altre nazioni arabe, lanciarono un attacco a sorpresa su Israele, su Yom Kippur. <hl> Questo rinnovo delle ostilità nel conflitto arabo-israeliano ha liberato la pressione economica sottostante sui prezzi del petrolio. All' epoca, l' Iran era il secondo esportatore mondiale di petrolio e un vicino alleato degli Stati Uniti. Settimane più tardi, lo scià d' Iran ha detto in un' intervista: Naturalmente[il prezzo del petrolio] sta andando a salire Certamente! E come! Avete[Paesi occidentali] aumentato il prezzo del grano che ci vendete del 300 per cento, e lo stesso per zucchero e cemento.",
        "<hl> Furono introdotti autocarri compatti, come la Toyota Hilux e il Datsun Truck, seguiti dal camion Mazda (venduto come il Ford Courier), e l' Isuzu costruito Chevrolet LUV. <hl> Mitsubishi rebranded il suo Forte come Dodge D-50 pochi anni dopo la crisi petrolifera. Mazda, Mitsubishi e Isuzu avevano partnership congiunte rispettivamente con Ford, Chrysler e GM. In seguito i produttori americani introdussero le loro sostituzioni nazionali (Ford Ranger, Dodge Dakota e la Chevrolet S10/GMC S-15), ponendo fine alla loro politica di importazione vincolata."
    ]
}
sample_lmqg_dict = {
    "en": ["William Turner was an English painter who specialised in watercolour landscapes", "William Turner"],
    "ja": ["フェルメールの作品では、17世紀のオランダの画家、ヨハネス・フェルメールの作品について記述する。フェルメールの作品は、疑問作も含め30数点しか現存しない。現存作品はすべて油彩画で、版画、下絵、素描などは残っていない。", "30数点"],
    "ru": ["Нелишним будет отметить, что, развивая это направление, Д. И. Менделеев, поначалу априорно выдвинув идею о температуре, при которой высота мениска будет нулевой, в мае 1860 года провёл серию опытов.", "в мае 1860 года"],
    "ko": ["1990년 영화 《 남부군 》에서 단역으로 영화배우 첫 데뷔에 이어 같은 해 KBS 드라마 《지구인》에서 단역으로 출연하였고 이듬해 MBC 《여명의 눈동자》를 통해 단역으로 출연하였다.", "남부군"],
    "es": ["a noviembre , que es también la estación lluviosa.", "noviembre"],
    "fr": ["Créateur » (Maker), lui aussi au singulier, « le Suprême Berger » (The Great Shepherd) ; de l'autre, des réminiscences de la théologie de l'Antiquité : le tonnerre, voix de Jupiter, « Et souvent ta voix gronde en un tonnerre terrifiant », etc.", "le Suprême Berger"],
    "de": ["das erste weltweit errichtete Hermann Brehmer 1855 im niederschlesischen ''Görbersdorf'' (heute Sokołowsko, Polen).", "1855"],
    "it": ["Dopo il 1971 , l' OPEC ha tardato ad adeguare i prezzi per riflettere tale deprezzamento.", "Dopo il 1971"]
}
language_dict = {
    "qg_frquad": 'fr',
    "qg_itquad": 'it',
    "qg_koquad": 'ko',
    "qg_dequad": 'de',
    "qg_jaquad": 'ja',
    "qg_ruquad": 'ru',
    "qg_esquad": 'es',
    "qag_frquad": 'fr',
    "qag_itquad": 'it',
    "qag_koquad": 'ko',
    "qag_dequad": 'de',
    "qag_jaquad": 'ja',
    "qag_ruquad": 'ru',
    "qag_esquad": 'es'
}


def keep_qag_metric(df):
    df = df.T
    col = df.columns
    for i in ["Bleu_1", "Bleu_2", "Bleu_3", "Bleu_4", "METEOR", "ROUGE_L", "BERTScore", "MoverScore"]:
        if i in col:
            df.pop(i)
    return df.T


def __format_metric(metric, metric_label, metric_label_type, is_multitask, is_end2end):
    tmp = ""
    if "QAAlignedF1Score (BERTScore)" not in metric['test']:
        if "Bleu_4" in metric["test"]:
            tmp += f"""
    - name: BLEU4 ({metric_label})
      type: bleu4_{metric_label_type}
      value: {metric["test"]["Bleu_4"]}"""
        if "ROUGE_L" in metric["test"]:
            tmp += f"""
    - name: ROUGE-L ({metric_label})
      type: rouge_l_{metric_label_type}
      value: {metric["test"]["ROUGE_L"]}"""
        if "METEOR" in metric["test"]:
            tmp += f"""
    - name: METEOR ({metric_label})
      type: meteor_{metric_label_type}
      value: {metric["test"]["METEOR"]}"""
        if "BERTScore" in metric["test"]:
            tmp += f"""
    - name: BERTScore ({metric_label})
      type: bertscore_{metric_label_type}
      value: {metric["test"]["BERTScore"]}"""
        if "MoverScore" in metric["test"]:
            tmp += f"""
    - name: MoverScore ({metric_label})
      type: moverscore_{metric_label_type}
      value: {metric["test"]["MoverScore"]}"""
        if "AnswerF1Score" in metric['test']:
            tmp += f"""
    - name: AnswerF1Score ({metric_label})
      type: answer_f1_score__{metric_label_type}
      value: {metric["test"]["AnswerF1Score"]}"""
        if "AnswerExactMatch" in metric['test']:
            tmp += f"""
    - name: AnswerExactMatch ({metric_label})
      type: answer_exact_match_{metric_label_type}
      value: {metric["test"]["AnswerExactMatch"]}"""
    if "QAAlignedF1Score (BERTScore)" in metric['test']:
        tmp += f"""
    - name: QAAlignedF1Score-BERTScore ({metric_label}){"" if is_multitask or is_end2end else " [Gold Answer]"}
      type: qa_aligned_f1_score_bertscore_{metric_label_type}{"" if is_multitask or is_end2end else "_gold_answer"}
      value: {metric["test"]["QAAlignedF1Score (BERTScore)"]}"""
    if "QAAlignedRecall (BERTScore)" in metric['test']:
        tmp += f"""
    - name: QAAlignedRecall-BERTScore ({metric_label}){"" if is_multitask or is_end2end else " [Gold Answer]"}
      type: qa_aligned_recall_bertscore_{metric_label_type}{"" if is_multitask or is_end2end else "_gold_answer"}
      value: {metric["test"]["QAAlignedRecall (BERTScore)"]}"""
    if "QAAlignedPrecision (BERTScore)" in metric['test']:
        tmp += f"""
    - name: QAAlignedPrecision-BERTScore ({metric_label}){"" if is_multitask or is_end2end else " [Gold Answer]"}
      type: qa_aligned_precision_bertscore_{metric_label_type}{"" if is_multitask or is_end2end else "_gold_answer"}
      value: {metric["test"]["QAAlignedPrecision (BERTScore)"]}"""
    if "QAAlignedF1Score (MoverScore)" in metric['test']:
        tmp += f"""
    - name: QAAlignedF1Score-MoverScore ({metric_label}){"" if is_multitask or is_end2end else " [Gold Answer]"}
      type: qa_aligned_f1_score_moverscore_{metric_label_type}{"" if is_multitask or is_end2end else "_gold_answer"}
      value: {metric["test"]["QAAlignedF1Score (MoverScore)"]}"""
    if "QAAlignedRecall (MoverScore)" in metric['test']:
        tmp += f"""
    - name: QAAlignedRecall-MoverScore ({metric_label}){"" if is_multitask or is_end2end else " [Gold Answer]"}
      type: qa_aligned_recall_moverscore_{metric_label_type}{"" if is_multitask or is_end2end else "_gold_answer"}
      value: {metric["test"]["QAAlignedRecall (MoverScore)"]}"""
    if "QAAlignedPrecision (MoverScore)" in metric['test']:
        tmp += f"""
    - name: QAAlignedPrecision-MoverScore ({metric_label}){"" if is_multitask or is_end2end else " [Gold Answer]"}
      type: qa_aligned_precision_moverscore_{metric_label_type}{"" if is_multitask or is_end2end else "_gold_answer"}
      value: {metric["test"]["QAAlignedPrecision (MoverScore)"]}"""
    return tmp


def format_metric(dataset, dataset_type, metric, metric_qag, metric_qa, metric_ae, metric_qag_pipe, is_multitask, is_end2end, is_qa, is_ae):
    metric_label = 'Question Generation'
    metric_label_type = "question_generation"
    if is_qa:
        metric_label = 'Question Answering'
        metric_label_type = "question_answering"
    elif is_ae:
        metric_label = 'Answer Extraction'
        metric_label_type = "answer_extraction"
    elif is_end2end:
        metric_label = 'Question & Answer Generation'
        metric_label_type = "question_answer_generation"
    tmp = f"""  - task:
      name: Text2text Generation
      type: text2text-generation
    dataset:
      name: {dataset}
      type: {dataset_type}
      args: {dataset_type}
    metrics:"""
    tmp += __format_metric(metric, metric_label, metric_label_type, is_multitask, is_end2end)
    if metric_qag is not None:
        tmp += __format_metric(metric_qag, 'Question & Answer Generation (with Gold Answer)', "question_answer_generation_with_gold_answer", is_multitask, is_end2end)
    if metric_qa:
        tmp += __format_metric(metric_qa, 'Question Answering', "question_answering", is_multitask, is_end2end)
    if metric_ae:
        tmp += __format_metric(metric_ae, 'Answer Extraction', "answer_extraction", is_multitask, is_end2end)
    if metric_qag_pipe:
        tmp += __format_metric(metric_qag_pipe, 'Question & Answer Generation', "question_answer_generation", is_multitask, is_end2end)
    return tmp


def format_usage(model_name, sample, sample_ae):
    if sample_ae is not None:
        return f"""from transformers import pipeline

pipe = pipeline("text2text-generation", "{model_name}")

# answer extraction
answer = pipe("{sample[0]}")

# question generation
question = pipe("{sample_ae[0]}")
"""
    else:
        return f"""from transformers import pipeline

pipe = pipeline("text2text-generation", "{model_name}")
output = pipe("{sample[0]}")
"""


def format_usage_lmqg(model_name, language, is_multitask, is_end2end, is_qa, is_ae):
    desc = f"""from lmqg import TransformersQG

# initialize model
model = TransformersQG(language="{language}", model="{model_name}")
"""
    if is_multitask or is_end2end:
        desc += f"""
# model prediction
question_answer_pairs = model.generate_qa("{sample_lmqg_dict[language][0]}")
"""
    elif is_qa:
        _tmp = sample_qa_dict[language][0]
        __q, __c = _tmp.split(', context:')
        __q = __q.replace("question: ", '')
        desc += f"""
# model prediction
answers = model.answer_q(list_question="{__q}", list_context="{__c}")
"""
    elif is_ae:
        desc += f"""
# model prediction
answers = model.generate_a("{sample_lmqg_dict[language][0]}")
"""
    else:
        desc += f"""
# model prediction
questions = model.generate_q(list_context="{sample_lmqg_dict[language][0]}", list_answer="{sample_lmqg_dict[language][1]}")
"""
    return desc


def get_readme(model_name: str, model_checkpoint: str):
    with open(pj(model_checkpoint, "trainer_config.json")) as f:
        config = json.load(f)
    config_text = "\n".join([f" - {k}: {v}" for k, v in config.items()])
    language_model = config['model']
    dataset = config['dataset_path']
    prefix_types = config['prefix_types']
    dataset_name = config['dataset_name']
    dataset_alias = os.path.basename(dataset)
    la = language_dict[dataset_alias] if dataset_alias in language_dict else 'en'
    header = f"This model is fine-tuned version of [{language_model}](https://huggingface.co/{language_model}) for question generation task on the [{dataset}](https://huggingface.co/datasets/{dataset}) (dataset_name: {dataset_name}) via [`lmqg`](https://github.com/asahi417/lm-question-generation)."
    metric_title = "Question Generation"
    # model_version
    eval_file = "metric.first.sentence.paragraph_answer.question"
    eval_file_qag = "metric.first.answer.paragraph.questions_answers"
    eval_file_qa = "metric.first.answer.paragraph_question.answer"
    eval_file_ae = "metric.first.answer.paragraph_sentence.answer"
    add_info = []
    _sample = sample_qg_dict[la]
    _is_qag = False
    _is_qa = False
    _is_ae = False
    _is_multitask = False

    # Multitask QAG Models
    if model_name.endswith('multitask') or '-qg-ae' in model_name:
        _is_multitask = True
        header = f"This model is fine-tuned version of [{language_model}](https://huggingface.co/{language_model}) for question generation and answer extraction jointly on the [{dataset}](https://huggingface.co/datasets/{dataset}) (dataset_name: {dataset_name}) via [`lmqg`](https://github.com/asahi417/lm-question-generation)."
    # E2E QAG Models
    elif 'qag' in model_name.split('-'):
        _sample = [re.sub(r'\s+', ' ', _sample[0].replace('<hl>', ''))]
        eval_file = "metric.first.answer.paragraph.questions_answers"
        metric_title = "Question & Answer Generation"
        eval_file_qag = eval_file_ae = None
        header = f"This model is fine-tuned version of [{language_model}](https://huggingface.co/{language_model}) for question & answer pair generation task on the [{dataset}](https://huggingface.co/datasets/{dataset}) (dataset_name: {dataset_name}) via [`lmqg`](https://github.com/asahi417/lm-question-generation)."
        _is_qag = True
        if 'np' in model_name.split('-'):
            add_info.append("This model is fine-tuned without a task prefix.")
    # QA Models
    elif "question-answering" in model_name or 'qa' in model_name.split('-'):
        _sample = sample_qa_dict[la]
        eval_file = "metric.first.answer.paragraph_question.answer"
        metric_title = "Question Answering"
        eval_file_qa = eval_file_qag = eval_file_ae = None
        header = f"This model is fine-tuned version of [{language_model}](https://huggingface.co/{language_model}) for question answering task on the [{dataset}](https://huggingface.co/datasets/{dataset}) (dataset_name: {dataset_name}) via [`lmqg`](https://github.com/asahi417/lm-question-generation)."
        _is_qa = True
    # Answer Extraction Models
    elif "answer-extraction" in model_name or ('ae' in model_name.split('-') and 'qg' not in model_name.split('-')):
        _sample = sample_ae_dict[la]
        eval_file = "metric.first.answer.paragraph_sentence.answer"
        metric_title = "Answer Extraction"
        eval_file_qa = eval_file_qag = eval_file_ae = None
        header = f"This model is fine-tuned version of [{language_model}](https://huggingface.co/{language_model}) for answer extraction on the [{dataset}](https://huggingface.co/datasets/{dataset}) (dataset_name: {dataset_name}) via [`lmqg`](https://github.com/asahi417/lm-question-generation)."
        _is_ae = True
    # QG Models
    else:
        if model_name.endswith('no-answer'):
            _sample = ["<hl> " + re.sub(r'\s+', ' ', i.replace('<hl>', '')) + " <hl>" for i in _sample]
            add_info.append(version_description['no-answer'])
            eval_file = "metric.first.sentence.paragraph_sentence.question"
        elif model_name.endswith('no-paragraph'):
            add_info.append(version_description['no-paragraph'])
            eval_file = "metric.first.sentence.sentence_answer.question"
        elif model_name.endswith('default'):
            add_info.append(version_description['default'])

    _sample = [re.sub(r"\A\s+", "", i) for i in _sample]
    add_info = ' '.join(add_info)

    # get widget
    _sample_ae = None
    if _is_qag:
        tags = "- questions and answers generation"
        _sample = _sample if prefix_types is None else [f'{TASK_PREFIX["qag"]}: {i}' for i in _sample]
        widget = '\n'.join([f"""- text: "{i}"\n  example_title: "Questions & Answers Generation Example {n + 1}" """ for n, i in enumerate(_sample)])
    elif _is_qa:
        tags = "- question answering"
        _sample = _sample if prefix_types is None else [f'{TASK_PREFIX["qa"]}: {i}' for i in _sample]
        widget = '\n'.join([f"""- text: "{i}"\n  example_title: "Question Answering Example {n + 1}" """ for n, i in enumerate(_sample)])
    elif _is_ae:
        tags = "- answer extraction"
        _sample = _sample if prefix_types is None else [f'{TASK_PREFIX["ae"]}: {i}' for i in _sample]
        widget = '\n'.join([f"""- text: "{i}"\n  example_title: "Answering Extraction Example {n + 1}" """ for n, i in enumerate(_sample)])
    else:
        if _is_multitask:  # multitask
            tags = "- question generation\n- answer extraction"
            _sample = [f'{TASK_PREFIX["qg"]}: {i}' for i in _sample]
            _sample_ae = [f'{TASK_PREFIX["ae"]}: {i}' for i in sample_ae_dict[la]]
            widget = '\n'.join([f"""- text: "{i}"\n  example_title: "Question Generation Example {n + 1}" """ for n, i in enumerate(_sample)])
            widget += '\n' + '\n'.join([f"""- text: "{i}"\n  example_title: "Answer Extraction Example {n + 1}" """ for n, i in enumerate(_sample_ae)])
        else:
            tags = "- question generation"
            _sample = _sample if prefix_types is None else [f'{TASK_PREFIX["qg"]}: {i}' for i in _sample]
            widget = '\n'.join([f"""- text: "{i}"\n  example_title: "Question Generation Example {n + 1}" """ for n, i in enumerate(_sample)])

    # usage
    usage = format_usage(model_name, _sample, _sample_ae)
    usage_lmqg = format_usage_lmqg(model_name, la, _is_multitask, _is_qag, _is_qa, _is_ae)

    # metric
    with open(pj(model_checkpoint, "eval", f"{eval_file}.{dataset.replace('/', '_')}.{dataset_name}.json")) as f:
        metric = {k: {_k: round(_v * 100 if _k not in ["AnswerF1Score", "AnswerExactMatch"] else _v, 2) for _k, _v in v.items()} for k, v in json.load(f).items()}

    metric_qag = None
    if eval_file_qag is not None:
        tmp_path = pj(model_checkpoint, "eval", f"{eval_file_qag}.{dataset.replace('/', '_')}.{dataset_name}.json")
        if os.path.exists(tmp_path):
            with open(tmp_path) as f:
                metric_qag = {k: {_k: round(_v * 100 if _k not in ["AnswerF1Score", "AnswerExactMatch"] else _v, 2) for _k, _v in v.items()} for k, v in json.load(f).items()}

    metric_qa = None
    if eval_file_qa is not None:
        tmp_path = pj(model_checkpoint, "eval", f"{eval_file_qa}.{dataset.replace('/', '_')}.{dataset_name}.json")
        if os.path.exists(tmp_path):
            with open(tmp_path) as f:
                metric_qa = {k: {_k: round(_v * 100 if _k not in ["AnswerF1Score", "AnswerExactMatch"] else _v, 2) for _k, _v in v.items()} for k, v in json.load(f).items()}

    metric_ae = None
    if eval_file_ae is not None:
        tmp_path = pj(model_checkpoint, "eval", f"{eval_file_ae}.{dataset.replace('/', '_')}.{dataset_name}.json")
        if os.path.exists(tmp_path):
            with open(tmp_path) as f:
                metric_ae = {k: {_k: round(_v * 100 if _k not in ["AnswerF1Score", "AnswerExactMatch"] else _v, 2) for _k, _v in v.items()} for k, v in json.load(f).items()}

    metric_qag_pipeline = None
    tmp_path = pj(model_checkpoint, "eval_pipeline", f"metric.first.answer.paragraph.questions_answers.{dataset.replace('/', '_')}.{dataset_name}.{model_name.replace('-qg', '-ae').replace('/', '_')}.json")
    if os.path.exists(tmp_path):
        with open(tmp_path) as f:
            metric_qag_pipeline = {k: {_k: round(_v * 100, 2) for _k, _v in v.items()} for k, v in json.load(f).items()}
    metric_main = [dataset, dataset_name, metric, metric_qag, metric_qa, metric_ae, metric_qag_pipeline]

    # metric for ood
    metrics_ood = []
    for i in sorted(glob(pj(model_checkpoint, "eval_ood", f"{eval_file}.*.json"))):
        ood_data = os.path.basename(i).split(f"{eval_file}.")[-1].replace('.json', '')
        _dataset, _dataset_name = ood_data.split(".")
        try:
            if _dataset_name == "default":
                load_dataset(_dataset)
            else:
                load_dataset(_dataset, _dataset_name)
        except FileNotFoundError:
            org = _dataset.split("_")[0]
            _dataset_main = '_'.join(_dataset.split("_")[1:])
            _dataset = f"{org}/{_dataset_main}"
            try:
                if _dataset_name == "default":
                    load_dataset(_dataset)
                else:
                    load_dataset(_dataset, _dataset_name)
            except FileNotFoundError:
                raise ValueError(f"dataset {_dataset} is not found")
        with open(i) as f:
            metric = json.load(f)
            metrics_ood.append([_dataset, _dataset_name, metric, None, None, None, None])

    metrics_text = '\n'.join([
        format_metric(
            dataset=d,
            dataset_type=t,
            metric=m,
            metric_qag=m_qag,
            metric_qa=m_qa,
            metric_ae=m_ae,
            metric_qag_pipe=m_qag_pipe,
            is_multitask=_is_multitask,
            is_end2end=_is_qag,
            is_qa=_is_qa,
            is_ae=_is_ae
        ) for d, t, m, m_qag, m_qa, m_ae, m_qag_pipe in [metric_main] + metrics_ood])
    # readme table
    df_main = pd.DataFrame(*list(zip(*list(metric_main[2]["test"].items())))[::-1], columns=["Score"])
    df_main['Type'] = metric_main[1]
    df_main['Dataset'] = f"[{metric_main[0]}](https://huggingface.co/datasets/{metric_main[0]})"
    link_main = f'https://huggingface.co/{model_name}/raw/main/eval/{eval_file}.{dataset.replace("/", "_")}.{dataset_name}.json'
    df_main = df_main.sort_index()
    if metric_title == "Question & Answer Generation":
        df_main = keep_qag_metric(df_main)
    markdown_table = f"""
- ***Metric ({metric_title})***: [raw metric file]({link_main}) 

{df_main.to_markdown()}

"""
    if metric_main[3] is not None:
        df_qag = pd.DataFrame(*list(zip(*list(metric_main[3]["test"].items())))[::-1], columns=["Score"])
        df_qag['Type'] = metric_main[1]
        df_qag['Dataset'] = f"[{metric_main[0]}](https://huggingface.co/datasets/{metric_main[0]})"
        link_qag = f'https://huggingface.co/{model_name}/raw/main/eval/{eval_file_qag}.{dataset.replace("/", "_")}.{dataset_name}.json'
        df_qag = df_qag.sort_index()
        df_qag = keep_qag_metric(df_qag)
        markdown_table += f"""
- ***Metric (Question & Answer Generation{', Reference Answer' if not _is_multitask else ''})***: {"" if _is_multitask else "Each question is generated from *the gold answer*."} [raw metric file]({link_qag})

{df_qag.to_markdown()}

"""
    if metric_main[6] is not None:
        df_qag_pipe = pd.DataFrame(*list(zip(*list(metric_main[6]["test"].items())))[::-1], columns=["Score"])
        df_qag_pipe['Type'] = metric_main[1]
        df_qag_pipe['Dataset'] = f"[{metric_main[0]}](https://huggingface.co/datasets/{metric_main[0]})"
        link_qag_pipe = f"https://huggingface.co/{model_name}/raw/main/eval_pipeline/metric.first.answer.paragraph.questions_answers.{dataset.replace('/', '_')}.{dataset_name}.{model_name.replace('-qg', '-ae').replace('/', '_')}.json"
        ae_model = f"https://huggingface.co/{model_name.replace('-qg', '-ae')}"
        df_qag_pipe = df_qag_pipe.sort_index()
        df_qag_pipe = keep_qag_metric(df_qag_pipe)
        markdown_table += f"""
- ***Metric (Question & Answer Generation, Pipeline Approach)***: Each question is generated on the answer generated by [`{model_name.replace('-qg', '-ae')}`]({ae_model}). [raw metric file]({link_qag_pipe})

{df_qag_pipe.to_markdown()}

"""
    if metric_main[4] is not None:
        df_qa = pd.DataFrame(*list(zip(*list(metric_main[4]["test"].items())))[::-1], columns=["Score"])
        df_qa['Type'] = metric_main[1]
        df_qa['Dataset'] = f"[{metric_main[0]}](https://huggingface.co/datasets/{metric_main[0]})"
        link_qa = f'https://huggingface.co/{model_name}/raw/main/eval/{eval_file_qa}.{dataset.replace("/", "_")}.{dataset_name}.json'
        df_qa = df_qa.sort_index()
        markdown_table += f"""
- ***Metric (Question Answering)***: [raw metric file]({link_qa})

{df_qa.to_markdown()}

"""
    if metric_main[5] is not None:
        df_ae = pd.DataFrame(*list(zip(*list(metric_main[5]["test"].items())))[::-1], columns=["Score"])
        df_ae['Type'] = metric_main[1]
        df_ae['Dataset'] = f"[{metric_main[0]}](https://huggingface.co/datasets/{metric_main[0]})"
        link_ae = f'https://huggingface.co/{model_name}/raw/main/eval/{eval_file_ae}.{dataset.replace("/", "_")}.{dataset_name}.json'
        df_ae = df_ae.sort_index()
        markdown_table += f"""
- ***Metric (Answer Extraction)***: [raw metric file]({link_ae})

{df_ae.to_markdown()}

"""
    if len(metrics_ood) != 0:
        content = "\n".join([
                      f"| [{d}](https://huggingface.co/datasets/{d}) | {t} | {round(100 * m['test']['BERTScore'], 2)} | {round(100 * m['test']['Bleu_4'], 2)} | {round(100 * m['test']['METEOR'], 2)} | {round(100 * m['test']['MoverScore'], 2)} | {round(100 * m['test']['ROUGE_L'], 2)} | "
                      f"[link](https://huggingface.co/{model_name}/raw/main/eval_ood/{eval_file}.{d.replace('/', '_')}.{t}.json) |"
                      for d, t, m, _, _, _, _ in metrics_ood])
        markdown_table += f"""
- ***Metrics ({metric_title}, Out-of-Domain)***
        
| Dataset | Type | BERTScore| Bleu_4 | METEOR | MoverScore | ROUGE_L | Link |
|:--------|:-----|---------:|-------:|-------:|-----------:|--------:|-----:|
{content}
"""
    return f"""
---
license: cc-by-4.0
metrics:
- bleu4
- meteor
- rouge-l
- bertscore
- moverscore
language: {la}
datasets:
- {dataset}
pipeline_tag: text2text-generation
tags:
{tags}
widget:
{widget}
model-index:
- name: {model_name}
  results:
{metrics_text}
---

# Model Card of `{model_name}`
{header}
{add_info}

### Overview
- **Language model:** [{language_model}](https://huggingface.co/{language_model})   
- **Language:** {la}  
- **Training data:** [{dataset}](https://huggingface.co/datasets/{dataset}) ({dataset_name})
- **Online Demo:** [https://autoqg.net/](https://autoqg.net/)
- **Repository:** [https://github.com/asahi417/lm-question-generation](https://github.com/asahi417/lm-question-generation)
- **Paper:** {paper_link}

### Usage
- With [`lmqg`](https://github.com/asahi417/lm-question-generation#lmqg-language-model-for-question-generation-)
```python
{usage_lmqg}
```

- With `transformers`
```python
{usage}
```

## Evaluation

{markdown_table}

## Training hyperparameters

The following hyperparameters were used during fine-tuning:
{config_text}

The full configuration can be found at [fine-tuning config file](https://huggingface.co/{model_name}/raw/main/trainer_config.json).

## Citation
```
{bib}
```
"""



""" T5 model. """
import os
import logging
import pickle
import re
import urllib
from itertools import chain
from typing import List, Dict
from multiprocessing import Pool

import numpy as np
from tqdm import tqdm
import torch
from torch.nn import functional
import transformers
from .exceptions import ExceedMaxLengthError, HighlightNotFoundError, AnswerNotFoundError
from .spacy_module import SpacyPipeline, VALID_METHODS

__all__ = ('TransformersQG', 'ADDITIONAL_SP_TOKENS', 'TASK_PREFIX', 'clean', 'internet_connection')

os.environ["TOKENIZERS_PARALLELISM"] = "false"  # to turn off warning message
TASK_PREFIX = {
    "ae": "extract answers",
    "qg": "generate question",
    "qag": "generate question and answer",
    "qa": "answer question"
}
CE_IGNORE_INDEX = -100
ADDITIONAL_SP_TOKENS = {'hl': '<hl>'}
NUM_WORKERS = int(os.getenv('NUM_WORKERS', '0'))
PARALLEL_PROCESSING = bool(int(os.getenv('PARALLEL_PROCESSING', '0')))
DEFAULT_MODELS = {
    'en': 'lmqg/t5-small-squad-qg-ae',
    'ja': 'lmqg/mt5-small-jaquad-qg-ae',
    'de': 'lmqg/mt5-small-dequad-qg-ae',
    'es': 'lmqg/mt5-small-esquad-qg-ae',
    'ko': 'lmqg/mt5-small-koquad-qg-ae',
    'ru': 'lmqg/mt5-small-ruquad-qg-ae',
    'it': 'lmqg/mt5-small-itquad-qg-ae',
    'fr': 'lmqg/mt5-small-frquad-qg-ae'
}


def pickle_save(obj, path: str):
    with open(path, "wb") as fp:
        pickle.dump(obj, fp)


def pickle_load(path: str):
    with open(path, "rb") as fp:  # Unpickling
        return pickle.load(fp)


def clean(string):
    string = re.sub(r'\A\s*', '', string)
    string = re.sub(r'\s*\Z', '', string)
    if len(string) > 0:
        return string
    return None


def internet_connection(host='http://google.com'):
    try:
        urllib.request.urlopen(host)
        return True
    except:
        return False


def load_language_model(model_name, cache_dir: str = None, use_auth_token: bool = False):
    """ load language model from huggingface model hub """
    # tokenizer
    local_files_only = not internet_connection()
    tokenizer = transformers.AutoTokenizer.from_pretrained(
        model_name, cache_dir=cache_dir, local_files_only=local_files_only, use_auth_token=use_auth_token)
    config = transformers.AutoConfig.from_pretrained(
        model_name, local_files_only=local_files_only, cache_dir=cache_dir, use_auth_token=use_auth_token)

    # model
    if config.model_type == 't5':  # T5 model requires T5ForConditionalGeneration class
        model_class = transformers.T5ForConditionalGeneration.from_pretrained
    elif config.model_type == 'mt5':
        model_class = transformers.MT5ForConditionalGeneration.from_pretrained
    elif config.model_type == 'bart':
        model_class = transformers.BartForConditionalGeneration.from_pretrained
    elif config.model_type == 'mbart':
        model_class = transformers.MBartForConditionalGeneration.from_pretrained
    else:
        raise ValueError(f'unsupported model type: {config.model_type}')
    model = model_class(
        model_name, config=config, cache_dir=cache_dir, local_files_only=local_files_only, use_auth_token=use_auth_token)
    # add new special tokens to the tokenizer and the model if they don't have it
    tokenizer.add_special_tokens({'additional_special_tokens': list(ADDITIONAL_SP_TOKENS.values())})
    model.resize_token_embeddings(len(tokenizer))
    return tokenizer, model, config


def label_smoothed_loss(logits, labels, epsilon):
    """ https://github.com/huggingface/transformers/blob/55bb4c06f7be141c6d895dbe1f11018dc8580b2d/src/transformers/trainer_pt_utils.py#L430 """
    log_probs = - functional.log_softmax(logits, dim=-1)
    if labels.dim() == log_probs.dim() - 1:
        labels = labels.unsqueeze(-1)

    padding_mask = labels.eq(CE_IGNORE_INDEX)
    # In case the ignore_index is -100, the gather will fail, so we replace labels by 0. The padding_mask
    # will ignore them in any case.
    labels.clamp_min_(0)

    nll_loss = log_probs.gather(dim=-1, index=labels)
    nll_loss.masked_fill_(padding_mask, 0.0)

    # works for fp16 input tensor too, by internally upcasting it to fp32
    smoothed_loss = log_probs.sum(dim=-1, keepdim=True, dtype=torch.float32)
    smoothed_loss.masked_fill_(padding_mask, 0.0)

    # Take the mean over the label dimensions, then divide by the number of active elements (i.e. not-padded):
    num_active_elements = padding_mask.numel() - padding_mask.long().sum()
    nll_loss = nll_loss.sum() / num_active_elements
    smoothed_loss = smoothed_loss.sum() / (num_active_elements * log_probs.shape[-1])
    return (1 - epsilon) * nll_loss + epsilon * smoothed_loss


class Dataset(torch.utils.data.Dataset):
    """ torch.utils.data.Dataset wrapper converting into tensor """
    float_tensors = ['attention_mask']

    def __init__(self, data: List):
        self.data = data

    def __len__(self):
        return len(self.data)

    def to_tensor(self, name, data):
        if name in self.float_tensors:
            return torch.tensor(data, dtype=torch.float32)
        return torch.tensor(data, dtype=torch.long)

    def __getitem__(self, idx):
        return {k: self.to_tensor(k, v) for k, v in self.data[idx].items()}


class EncodePlus:
    """ Wrapper of encode_plus for multiprocessing. """

    def __init__(self,
                 tokenizer,
                 max_length: int = 512,
                 max_length_output: int = 34,
                 drop_overflow_error_text: bool = False,
                 skip_overflow_error: bool = False,
                 drop_highlight_error_text: bool = False,
                 prefix_type: str = None,
                 padding: bool = True):
        """ Wrapper of encode_plus for multiprocessing.

        @param tokenizer: transforms.Tokenizer
        @param max_length: Max text length of input.
        @param max_length_output: Max text length of output.
        @param drop_overflow_error_text: If true, return None when the input exceeds the max length.
        @param skip_overflow_error: If true, raise an error when the input exceeds the max length.
        @param drop_highlight_error_text: If true, raise an error when a highlight span is not found in the paragraph.
        @param prefix_type: Either of `qg` or `answer_extraction`, which is to add at the beginning of the text.
        @param padding: Pad the sequence to the max length.
        """
        self.prefix = TASK_PREFIX[prefix_type] if prefix_type is not None else None
        self.tokenizer = tokenizer
        self.max_length = max_length
        self.max_length_output = max_length_output
        # NOTE: for model training, we should drop the exceeded input but not for the evaluation
        self.drop_overflow_error_text = drop_overflow_error_text
        self.skip_overflow_error = skip_overflow_error
        self.drop_highlight_error_text = drop_highlight_error_text
        # truncation should be true for the batch process, but not necessary to process single input
        self.param_in = {'truncation': True, 'max_length': self.max_length}
        self.param_out = {'truncation': True, 'max_length': self.max_length_output}
        if padding:
            self.param_in['padding'] = 'max_length'
            self.param_out['padding'] = 'max_length'

    def __call__(self, inputs):
        return self.encode_plus(*inputs)

    def encode_plus(self, input_sequence: str, output_sequence: str = None, input_highlight: str = None):
        """ encode_plus

        @param input_sequence: Input sequence.
        @param output_sequence: Output sequence.
        @param input_highlight: Sub-sequence of `input_sequence` to be surrounded by <hl>.
        @return: The output of `encode_plus`.
        """

        # add highlight to the input
        if input_highlight is not None:
            position = input_sequence.find(input_highlight)
            if position == -1:
                if self.drop_highlight_error_text:
                    return None
                raise HighlightNotFoundError(input_highlight, input_sequence)
            input_sequence = '{0}{1} {2} {1}{3}'.format(
                input_sequence[:position], ADDITIONAL_SP_TOKENS['hl'], input_highlight,
                input_sequence[position+len(input_highlight):])

        if self.prefix is not None:
            input_sequence = f'{self.prefix}: {input_sequence}'

        # handling overflow text
        # drop_overflow_error_text ==> remove the overflow sentence from input
        # skip_overflow_error ==> keep the overflow sentence
        # none of them ==> raise error
        if self.drop_overflow_error_text or not self.skip_overflow_error:
            if len(self.tokenizer.encode(input_sequence)) > self.max_length:
                if not self.drop_overflow_error_text:  # raise error for overflow text
                    raise ExceedMaxLengthError(self.max_length)
                return None  # remove overflow text
            if output_sequence is not None:
                if len(self.tokenizer.encode(output_sequence)) > self.max_length_output:
                    if not self.drop_overflow_error_text:  # raise error for overflow text
                        raise ExceedMaxLengthError(self.max_length)
                    return None  # remove overflow text
        encode = self.tokenizer.encode_plus(input_sequence, **self.param_in)
        if output_sequence is not None:
            encode['labels'] = self.tokenizer.encode(output_sequence, **self.param_out)
        return encode


class TransformersQG:
    """ Transformers Language Model for Question Generation. """

    def __init__(self,
                 model: str = None,
                 max_length: int = 512,
                 max_length_output: int = 256,
                 model_ae: str = None,
                 max_length_ae: int = 512,
                 max_length_output_ae: int = 64,
                 cache_dir: str = None,
                 add_prefix: bool = None,
                 language: str = 'en',
                 label_smoothing: float = None,
                 skip_overflow_error: bool = False,
                 drop_overflow_error_text: bool = False,
                 drop_highlight_error_text: bool = False,
                 drop_answer_error_text: bool = False,
                 use_auth_token: bool = False,
                 is_qg: bool = None,
                 is_qag: bool = None,
                 is_qa: bool = None,
                 is_ae: bool = None):
        """ Transformers Language Model for Question Generation.

        @param model: Model alias or path to local model file.
        @param max_length: Max text length of input.
        @param max_length_output: Max text length of output.
        @param cache_dir: Directory to cache transformers model files.
        @param add_prefix: Whether model uses task-specific prefix (eg. True for T5 but False for BART models).
        @param language: Language alias for SpaCy language-specific pipelines (sentencizer/keyword extraction).
        @param label_smoothing: [Fine-tuning parameter] Label smoothing.
        @param drop_overflow_error_text: If true, return None when the input exceeds the max length.
        @param skip_overflow_error: If true, raise an error when the input exceeds the max length.
        @param drop_highlight_error_text: If true, raise an error when a highlight span is not found in the paragraph.
        @param use_auth_token: [optional] Huggingface transformers argument of `use_auth_token`
        """

        # take default model given the language
        if model is None:
            assert language in DEFAULT_MODELS.keys(),\
                f"Model with language '{language}' is not available. Please choose language from " \
                f"'{DEFAULT_MODELS.keys()}' or specify 'model'."
            model = DEFAULT_MODELS[language]

        # classify model type
        self.is_qg = 'qg' in model.split('-') if is_qg is None else is_qg
        self.is_ae = 'ae' in model.split('-') if is_ae is None else is_ae
        self.is_qa = 'qa' in model.split('-') if is_qa is None else is_qa
        self.is_qag = 'qag' in model.split('-') if is_qag is None else is_qag

        # configs
        self.model_name = model
        self.max_length = max_length
        self.max_length_output = max_length_output
        self.label_smoothing = label_smoothing
        self.drop_overflow_error_text = drop_overflow_error_text
        self.skip_overflow_error = skip_overflow_error
        self.drop_highlight_error_text = drop_highlight_error_text
        self.drop_answer_error_text = drop_answer_error_text
        self.model_name_ae = model_ae
        self.max_length_ae = max_length_ae
        self.max_length_output_ae = max_length_output_ae
        # load model
        self.tokenizer, self.model, config = load_language_model(self.model_name, cache_dir=cache_dir, use_auth_token=use_auth_token)
        if 'add_prefix' not in config.to_dict().keys():
            # this means the model is not fine-tuned
            # assert add_prefix, '`add_prefix` is required for non-fine-tuned models'
            self.add_prefix = add_prefix
        else:
            self.add_prefix = config.add_prefix

        # set default behaviour for answer extraction
        if self.model_name_ae is None:
            self.model_name_ae = self.model_name if self.is_ae else "positionrank"
        # load answer extraction model
        self.answer_model_type = None
        if self.model_name_ae in VALID_METHODS:
            logging.info(f'use spaCy answer extraction model: {self.model_name_ae}')
            self.tokenizer_ae = self.model_ae = self.add_prefix_ae = None
            self.spacy_module = SpacyPipeline(language, self.model_name_ae)
            self.answer_model_type = 'spacy'
        else:
            logging.info(f'use LMQG fine-tuned answer extraction model: {self.model_name_ae}')
            if self.model_name == self.model_name_ae:
                logging.info("the same model as QG is used as AE")
                assert self.is_ae, f"the model ({self.model_name_ae}) is not fine-tuned for AE"
                self.tokenizer_ae = self.model_ae = self.add_prefix_ae = None
                self.answer_model_type = 'multitask'
            else:
                logging.info(f"loading 2nd model for AE: {self.model_name_ae}")
                self.tokenizer_ae, self.model_ae, config_ae = load_language_model(model_ae, cache_dir=cache_dir, use_auth_token=use_auth_token)
                self.add_prefix_ae = config_ae.add_prefix
                self.answer_model_type = 'pipeline'
            self.spacy_module = SpacyPipeline(language)

        # GPU setup
        self.device = 'cuda' if torch.cuda.device_count() > 0 else 'cpu'
        self.parallel = False
        if torch.cuda.device_count() > 1:
            self.parallel = True
            self.model = torch.nn.DataParallel(self.model)
            if self.model_ae is not None:
                self.model_ae = torch.nn.DataParallel(self.model_ae)
        self.model.to(self.device)
        if self.model_ae is not None:
            self.model_ae.to(self.device)
        logging.info(f'Model `{self.model_name}`')
        logging.info(f'\t * Num of GPU in use: {torch.cuda.device_count()}')
        logging.info(f'\t * Prefix: {self.add_prefix}')
        logging.info(f'\t * Language: {language} (ignore at the training phase)')

    def generate_qa_end2end(self,
                            list_context: str or List,
                            batch_size: int = None,
                            num_beams: int = 4,
                            cache_path: str = None,
                            splitting_symbol: str = '|',
                            question_prefix: str = "question: ",
                            answer_prefix: str = ", answer: "):
        """ Generate question from paragraph and answer. Note that `list_answer` is needed unless they are already
        highlighted in the `list_context`. eg) "I live in <hl> Tokyo <hl>."

        @param list_context: List of input texts.
        @param batch_size: Batch size.
        @param num_beams: Number of beam for model generation.
        @param cache_path: Path to pre-compute features.
        @return: List of generated sentences.
        """
        logging.info(f'running model for `question_answer_pair_generation`')
        assert self.is_qag, "`generate_qa_end2end` is available for end2end_qag_model"
        prefix_type = 'qag' if self.add_prefix else None
        single_input = type(list_context) is str
        list_context = [list_context] if single_input else list_context
        output = self.generate_prediction(
            list_context, prefix_type=prefix_type, cache_path=cache_path, num_beams=num_beams, batch_size=batch_size
        )

        def format_qa(list_raw_string):
            tmp = []
            for raw_string in list_raw_string:
                if len(raw_string.split(answer_prefix)) != 2 or question_prefix not in raw_string:
                    logging.info(f"invalid prediction: {raw_string}")
                else:
                    q, a = raw_string.split(answer_prefix)
                    a = re.sub(r'\A\s+', '', a)
                    a = re.sub(r'\s+\Z', '', a)
                    q = q.replace(question_prefix, "")
                    q = re.sub(r'\A\s+', '', q)
                    q = re.sub(r'\s+\Z', '', q)
                    tmp.append((q, a))
            return tmp

        output = [format_qa(o.split(splitting_symbol)) for o in output]
        return output[0] if single_input else output

    def generate_qa(self,
                    list_context: str or List,
                    batch_size: int = None,
                    num_beams: int = 4,
                    cache_path: str = None,
                    num_questions: int = None,
                    sentence_level: bool = False):
        """ Generate question given context.

        @param list_context: Input text.
        @param batch_size: Batch size.
        @param num_beams: Number of beam for model generation.
        @param cache_path: Path to pre-compute features.
        @param num_questions: Max number of questions.
        @param sentence_level: Run prediction on each sentence of the context independently to reduce complexity.
        @return: List of generated sentences.
        """
        if self.is_qag:
            return self.generate_qa_end2end(list_context, batch_size, num_beams, cache_path)
        single_input = type(list_context) is str
        list_context = [list_context] if single_input else list_context
        original_input_length = len(list_context)

        logging.info('running model for `ae`')
        list_answer = self.generate_a(
            list_context,
            batch_size=batch_size,
            num_beams=num_beams,
            cache_path=cache_path,
            sentence_level=sentence_level,
            num_questions=num_questions
        )
        valid_context_id = [n for n, a in enumerate(list_answer) if a is not None]
        list_context = [list_context[n] for n in valid_context_id]
        list_answer = [list_answer[n] for n in valid_context_id]
        qg_input, qg_hl, list_length = [], [], [0]
        for c, a in zip(list_context, list_answer):
            qg_hl += a
            qg_input += [c] * len(a)
            list_length.append(list_length[-1] + len(a))

        logging.info('running model for `qg`')
        list_question = self.generate_q(
            qg_input,
            list_answer=qg_hl,
            batch_size=batch_size,
            cache_path=cache_path,
            num_beams=num_beams,
            sentence_level=sentence_level
        )

        assert len(qg_hl) == len(list_question), f"{len(qg_input)} != {len(list_question)}"

        # return to nested list
        list_question = [list_question[list_length[n - 1]:list_length[n]] for n in range(1, len(list_length))]
        list_answer = [qg_hl[list_length[n - 1]:list_length[n]] for n in range(1, len(list_length))]
        output_list = [None] * original_input_length
        # print(len(valid_context_id), valid_context_id[:10], valid_context_id[-10:0])
        # print(original_input_length)
        # print(len(list_question), len(list_answer))
        for n, _id in enumerate(valid_context_id):
            output_list[_id] = [(q, a) for q, a in zip(list_question[n], list_answer[n])]
        return output_list[0] if single_input else output_list

    def generate_a(self,
                   context: str or List,
                   batch_size: int = None,
                   num_beams: int = 4,
                   cache_path: str = None,
                   sentence_level: bool = False,
                   num_questions: int = None):
        """ Generate answers from each sentence.

        @param context: Input text.
        @param batch_size: Batch size.
        @param num_beams: Number of beam for model generation.
        @param cache_path: Path to pre-compute features.
        @param sentence_level: Run prediction on each sentence of the context independently to reduce complexity.
        @param num_questions: Max number of questions.
        @return: List of generated answers.
        """
        logging.info(f'running model for `answer_extraction`')
        if self.answer_model_type == 'spacy':
            num_questions = 10 if num_questions is None else num_questions
            if type(context) is str:
                return self.spacy_module.keyword(context, num_questions)
            else:
                return [self.spacy_module.keyword(c, num_questions) for c in context]
        single_input = type(context) is str
        context = [context] if single_input else context
        list_sentences = [self.spacy_module.sentence(c) for c in context]  # split into sentence
        list_inputs = [[c] * len(s) for c, s in zip(context, list_sentences)]
        list_length = [0] + np.cumsum([len(s) for s in list_sentences]).tolist()
        if sentence_level:
            list_inputs = list_sentences
        # flatten inputs
        flat_sentences = list(chain(*list_sentences))
        flat_inputs = list(chain(*list_inputs))
        if self.answer_model_type == 'multitask':
            answer = self.generate_prediction(
                flat_inputs,  # list_input,
                highlights=flat_sentences,  # highlights=list_sentence,
                prefix_type='ae' if self.add_prefix else None,
                cache_path=cache_path,
                num_beams=num_beams,
                batch_size=batch_size
            )
        elif self.answer_model_type == 'pipeline':
            answer = self.generate_prediction(
                flat_inputs,  # list_input,
                highlights=flat_sentences,  # highlights=list_sentence,
                prefix_type='ae' if self.add_prefix_ae else None,
                cache_path=cache_path,
                num_beams=num_beams,
                batch_size=batch_size,
                switch_to_model_ae=True
            )
        else:
            raise ValueError(f"unknown answer model type: {self.answer_model_type}")
        # return to nested list
        answer = [clean(a) for a in answer]
        list_answer = [answer[list_length[n - 1]:list_length[n]] for n in range(1, len(list_length))]
        list_answer = [[a for a, c in zip(a_sent, c_sent) if a is not None and a in c]
                       for a_sent, c_sent in zip(list_answer, list_inputs)]
        list_answer = [None if len(a) == 0 else a for a in list_answer]
        if not self.drop_answer_error_text:
            if any(a is None for a in list_answer):
                raise AnswerNotFoundError([context[n] for n, a in enumerate(list_answer) if a is None][0])
        return list_answer[0] if single_input else list_answer

    def generate_q(self,
                   list_context: str or List,
                   list_answer: List = None,
                   batch_size: int = None,
                   num_beams: int = 4,
                   cache_path: str = None,
                   sentence_level: bool = False):
        """ Generate question from paragraph and answer. Note that `list_answer` is needed unless they are already
        highlighted in the `list_context`. eg) "I live in <hl> Tokyo <hl>."

        @param list_context: List of input texts.
        @param list_answer: List of answers in the `list_context` that are highlighted by <hl>.
        @param batch_size: Batch size.
        @param num_beams: Number of beam for model generation.
        @param cache_path: Path to pre-compute features.
        @param sentence_level: Run prediction on each sentence of the context independently to reduce complexity.
        @return: List of generated sentences.
        """
        assert self.is_qg, "model is not fine-tuned for QG"
        if list_answer is not None:
            assert type(list_context) is type(list_answer), f"{type(list_context)} != {type(list_answer)}"
        single_input = False
        if type(list_context) is str:
            list_context = [list_context]
            list_answer = [list_answer] if list_answer is not None else None
            single_input = True
        output = self.generate_prediction(
            list_context,
            highlights=list_answer,
            prefix_type='qg' if self.add_prefix else None,
            cache_path=cache_path,
            num_beams=num_beams,
            batch_size=batch_size,
            sentence_level=sentence_level
        )
        if single_input:
            return output[0]
        return output

    def answer_q(self,
                 list_context: str or List,
                 list_question: str or List,
                 batch_size: int = None,
                 num_beams: int = 4,
                 cache_path: str = None):
        logging.info(f'running model for `question_answering`')
        assert self.is_qa, "model is not fine-tuned for QA"
        assert type(list_context) is type(list_question), "invalid input"
        single_input = type(list_context) is str
        list_context = [list_context] if single_input else list_context
        list_question = [list_question] if single_input else list_question
        assert len(list_context) == len(list_question), f"invalid input: {len(list_context)} != {len(list_question)}"
        output = self.generate_prediction(
            [f"question: {q}, context: {c}" for q, c in zip(list_question, list_context)],
            batch_size=batch_size,
            prefix_type='qa' if self.add_prefix else None,
            cache_path=cache_path,
            num_beams=num_beams
        )
        return output[0] if single_input else output

    def generate_prediction(self,
                            inputs: List,
                            highlights: List or None = None,
                            prefix_type: str = None,
                            num_beams: int = 4,
                            batch_size: int = None,
                            cache_path: str = None,
                            sentence_level: bool = False,
                            switch_to_model_ae: bool = False):
        """ General method to generate model prediction

        @param inputs: List of input sequences.
        @param highlights: List of sub-sequences from list_context to be highlighted by <hl>.
        @param batch_size: Batch size.
        @param num_beams: Number of beam for model generation.
        @param cache_path: Path to pre-compute features.
        @param prefix_type: Either of `qg` or `answer_extraction`, which is to add at the beginning of the text.
        @return: List of generated sequences.
        """
        self.eval()
        if switch_to_model_ae:
            assert self.model_ae is not None and self.tokenizer_ae is not None
            model = self.model_ae
            tokenizer = self.tokenizer_ae
            max_length_output = self.max_length_output_ae
        else:
            model = self.model
            tokenizer = self.tokenizer
            max_length_output = self.max_length_output

        if sentence_level:
            assert highlights is not None, '`sentence_level` needs `highlights`.'
            assert len(highlights) == len(inputs), str([len(highlights), len(inputs)])
            list_sentence = []
            for context, answer in zip(inputs, highlights):
                s = [sentence for sentence in self.spacy_module.sentence(context) if answer in sentence]
                list_sentence.append(s[0] if len(s) != 0 else context)
            inputs = list_sentence

        assert type(inputs) is list, inputs
        encode_list = self.text_to_encode(
            inputs,
            highlights=highlights,
            prefix_type=prefix_type,
            cache_path=cache_path,
            switch_to_model_ae=switch_to_model_ae
        )
        loader = self.get_data_loader(encode_list, batch_size=batch_size)
        outputs = []
        for encode in loader:
            with torch.no_grad():
                if 'labels' in encode:
                    encode.pop('labels')
                encode = {k: v.to(self.device) for k, v in encode.items()}
                encode['max_length'] = max_length_output
                encode['num_beams'] = num_beams
                tensor = model.module.generate(**encode) if self.parallel else model.generate(**encode)
                outputs += tokenizer.batch_decode(tensor, skip_special_tokens=True)
        return outputs

    def encode_to_loss(self, encode: Dict):
        """ Transform encoded features to loss value for model finetuning.

        @param encode: Encoded feature.
        @return: Loss value.
        """
        assert 'labels' in encode
        output = self.model(**{k: v.to(self.device) for k, v in encode.items()})
        if self.label_smoothing is None or self.label_smoothing == 0.0:
            return output['loss'].mean() if self.parallel else output['loss']
        else:
            return label_smoothed_loss(output['logits'], encode['labels'].to(self.device), self.label_smoothing)

    def text_to_encode(self,
                       inputs,
                       outputs: List = None,
                       highlights: List = None,
                       prefix_type: str = None,
                       cache_path: str = None,
                       switch_to_model_ae: bool = False):
        """ Transform texts into encoded features.

        @param inputs: List of input sequences.
        @param outputs: List of output sequences.
        @param highlights: List of sub-sequences from `inputs` to be highlighted by <hl>.
        @param prefix_type: Either of `qg` or `answer_extraction`, which is to add at the beginning of the text.
        @param cache_path: Path to pre-compute features.
        @return: List of encoded feature.
        """
        if cache_path is not None and os.path.exists(cache_path):
            logging.info(f'loading preprocessed feature from {cache_path}')
            return pickle_load(cache_path)
        outputs = [None] * len(inputs) if outputs is None else outputs
        highlights = [None] * len(inputs) if highlights is None else highlights
        assert len(outputs) == len(inputs) == len(highlights), str([len(outputs), len(inputs), len(highlights)])
        data = list(zip(inputs, outputs, highlights))

        # process in parallel/single
        config = {'tokenizer': self.tokenizer, 'max_length': self.max_length, 'prefix_type': prefix_type,
                  'max_length_output': self.max_length_output, 'drop_overflow_error_text': self.drop_overflow_error_text,
                  'skip_overflow_error': self.skip_overflow_error, 'drop_highlight_error_text': self.drop_highlight_error_text,
                  'padding': False if len(data) == 1 else True}
        if switch_to_model_ae:
            assert self.model_ae is not None and self.tokenizer_ae is not None
            config['tokenizer'] = self.tokenizer_ae
            config['max_length'] = self.max_length_ae
            config['max_length_output'] = self.max_length_output_ae

        logging.info(f'encode all the data       : {len(data)}')
        if cache_path is not None:
            os.makedirs(os.path.dirname(cache_path), exist_ok=True)
        if PARALLEL_PROCESSING:
            pool = Pool()
            out = pool.map(EncodePlus(**config), data)
            pool.close()
            out = list(filter(None, out))  # remove overflow text
        else:
            f = EncodePlus(**config)
            out = []
            files = []
            for i in tqdm(data):
                e = f(i)
                if e is not None:  # remove overflow text
                    out.append(e)
                if len(out) > 40000 and cache_path is not None:
                    pickle_save(out, f'{cache_path}.tmp{len(files)}')
                    files.append(f'{cache_path}.tmp{len(files)}')
                    out = []
            if len(out) > 0 and cache_path is not None:
                pickle_save(out, f'{cache_path}.tmp{len(files)}')
                files.append(f'{cache_path}.tmp{len(files)}')
            if len(files) > 0:
                out = list(chain(*[pickle_load(i) for i in files]))
        logging.info(f'after remove the overflow : {len(out)}')
        # cache the encoded data
        if cache_path is not None:
            pickle_save(out, cache_path)
            logging.info(f'preprocessed feature is saved at {cache_path}')
        return out

    def save(self, save_dir):
        """ Save model.

        @param save_dir: Directory to save model related file.
        """

        def model_state(model):
            if self.parallel:
                return model.module
            return model

        logging.info('saving model')
        model_state(self.model).config.update({'add_prefix': self.add_prefix})
        model_state(self.model).save_pretrained(save_dir)
        logging.info('saving tokenizer')
        self.tokenizer.save_pretrained(save_dir)

    @staticmethod
    def get_data_loader(encode_list, batch_size: int = None, shuffle: bool = False, drop_last: bool = False):
        """ Get torch.utils.data.DataLoader instance.

        @param encode_list: List of encoded features.
        @param batch_size: Batch size.
        @param shuffle: Shuffle data.
        @param drop_last: Drop residual batch.
        @return: torch.utils.data.DataLoader
        """
        batch_size = len(encode_list) if batch_size is None else batch_size
        params = dict(batch_size=batch_size, shuffle=shuffle, drop_last=drop_last, num_workers=NUM_WORKERS)
        return torch.utils.data.DataLoader(Dataset(encode_list), **params)

    def train(self):
        self.model.train()

    def eval(self):
        self.model.eval()



import logging
import os
import json
from os.path import join as pj
from tqdm import tqdm
from datasets import load_dataset
from ..language_model import TransformersQG


def generate_qa_pairs(
        model_qg: str = 'lmqg/t5-small-squad-qg-ae',
        model_ae: str = None,
        language: str = 'en',
        anchor_data: str = 'lmqg/qa_squadshifts',
        anchor_data_name: str = 'new_wiki',
        use_reference_answer: bool = False,
        batch_size: int = 256,
        export_dir: str = None,
        overwrite: bool = False,
        max_length: int = 512,
        max_length_output: int = 256):

    logging.info(f'generate QA pairs from {anchor_data} with {model_qg}')
    data = load_dataset(anchor_data) if anchor_data_name is None else load_dataset(anchor_data, anchor_data_name)

    if export_dir is not None:
        os.makedirs(export_dir, exist_ok=True)

    model = TransformersQG(
        model=model_qg,
        model_ae=model_ae,
        language=language,
        skip_overflow_error=True,
        drop_answer_error_text=True,
        max_length=max_length,
        max_length_output=max_length_output)

    logging.info(f"Export dir: {export_dir}")
    full_output = {}
    for _split in data:

        logging.info(f'running prediction on {_split}')
        if export_dir is not None:
            if not overwrite and os.path.exists(pj(export_dir, f'{_split}.jsonl')):
                with open(pj(export_dir, f'{_split}.jsonl')) as f:
                    full_output[_split] = [json.loads(i) for i in f.read().split('\n') if len(i) > 0]
                continue

        if use_reference_answer:
            question = model.generate_q(
                list_context=data[_split]['context'],
                list_answer=[a['text'][0] for a in data[_split]['answers']],
                batch_size=batch_size
            )
            if question is None or len(question) == 0:
                continue
            assert len(data[_split]) == len(question), f"{len(data[_split])} != {len(question)}"
            output = [{
                'id': tmp_data['id'],
                'title': tmp_data['title'],
                'context': tmp_data['context'],
                'question': q,
                'answers': tmp_data['answers']
            } for tmp_data, q in zip(data[_split], question)]

        else:
            list_context = []
            list_data = []
            for c, df in tqdm(data[_split].to_pandas().groupby('context')):
                list_context.append(c)
                list_data.append(df)
            qa_pairs = model.generate_qa(
                list_context=list_context,
                batch_size=batch_size
            )
            output = []
            _id = 0
            for _qa_pairs, _context, df in zip(qa_pairs, list_context, list_data):
                if _qa_pairs is None:
                    continue
                for q, a in _qa_pairs:
                    context = df['context'].values[0]
                    if a not in context:
                        continue
                    _id += 1
                    output.append({
                        'id': str(_id),
                        'title': df['title'].values[0],
                        'context': context,
                        'question': q,
                        'answers': {'text': [a], 'answer_start': [context.index(a)]}
                    })

        full_output[_split] = output
        if export_dir is not None:
            logging.info(f"saving {_split} at {pj(export_dir, f'{_split}.jsonl')}")
            with open(pj(export_dir, f'{_split}.jsonl'), 'w') as f:
                f.write('\n'.join([json.dumps(i) for i in output]))

    return full_output



""" Training model. """
import os
import json
import logging
import shutil
import random
from os.path import join as pj
from glob import glob
from typing import List
import torch

from .language_model import TransformersQG
from .data import get_dataset, DEFAULT_CACHE_DIR

__all__ = ('to_list', 'Trainer')

OPTIMIZER_ON_CPU = bool(int(os.getenv('OPTIMIZER_ON_CPU', '0')))


def to_list(_val, sorting=True):
    if type(_val) != list:
        return [_val]
    if sorting:
        return sorted(_val, reverse=True)
    return _val


class Config:
    """ Model checkpoint managing class. """

    def __init__(self, checkpoint_dir: str, config_file: str = 'trainer_config.json', **kwargs):
        self.checkpoint_dir = checkpoint_dir
        if os.path.exists(self.checkpoint_dir):
            logging.info(f'load config from existing checkpoint at {self.checkpoint_dir}')
            self.config = self.safe_open(pj(self.checkpoint_dir, config_file))
        else:
            logging.info(f'initialize checkpoint at {self.checkpoint_dir}')
            self.config = kwargs
            configs = {i: self.safe_open(i) for i in glob(pj(os.path.dirname(self.checkpoint_dir), '*', config_file))}
            configs = list(filter(lambda x: x[1] == self.config, configs.items()))
            if len(configs) != 0:
                input(f'\ncheckpoint with same config already exists: {configs[0]}\n enter to overwrite >>>')
                for _p, _ in configs:
                    shutil.rmtree(os.path.dirname(_p))
            self.__initialize_checkpoint(config_file)

        self.__dict__.update(self.config)
        logging.info('hyperparameters')
        for k, v in self.config.items():
            logging.info(f'\t * {k}: {str(v)[:min(100, len(str(v)))]}')

    def __initialize_checkpoint(self, config_file):
        os.makedirs(self.checkpoint_dir, exist_ok=True)
        if not os.path.exists(pj(self.checkpoint_dir, config_file)):
            with open(pj(self.checkpoint_dir, config_file), 'w') as f:
                json.dump(self.config, f)

    @staticmethod
    def safe_open(_file):
        with open(_file, 'r') as f:
            return json.load(f)


class Trainer:

    def __init__(self,
                 checkpoint_dir: str,
                 dataset_path: str = "lmqg/qg_squad",
                 dataset_name: str = 'default',
                 input_types: List or str = 'paragraph_answer',
                 output_types: List or str = 'question',
                 prefix_types: List or str = 'qg',
                 model: str = 't5-small',
                 max_length: int = 512,
                 max_length_output: int = 32,
                 epoch: int = 10,
                 batch: int = 128,
                 lr: float = 1e-4,
                 fp16: bool = False,
                 random_seed: int = 42,
                 gradient_accumulation_steps: int = 4,
                 label_smoothing: float = None,
                 disable_log: bool = False,
                 config_file: str = 'trainer_config.json',
                 use_auth_token: bool = False):
        logging.info('initialize model trainer')
        self.use_auth_token = use_auth_token
        # config
        self.config = Config(
            config_file=config_file, checkpoint_dir=checkpoint_dir, dataset_path=dataset_path, dataset_name=dataset_name,
            input_types=input_types, output_types=output_types, prefix_types=prefix_types, model=model,
            max_length=max_length, max_length_output=max_length_output, epoch=epoch, batch=batch, lr=lr, fp16=fp16,
            random_seed=random_seed, gradient_accumulation_steps=gradient_accumulation_steps,
            label_smoothing=label_smoothing)

        random.seed(self.config.random_seed)
        torch.manual_seed(self.config.random_seed)
        if not disable_log:
            # add file handler
            logger = logging.getLogger()
            file_handler = logging.FileHandler(pj(self.config.checkpoint_dir, 'training.log'))
            file_handler.setLevel(logging.DEBUG)
            file_handler.setFormatter(logging.Formatter('%(asctime)s %(levelname)-8s %(message)s'))
            logger.addHandler(file_handler)

        # load model
        add_prefix = False if self.config.prefix_types is None else True
        ckpts = glob(pj(self.config.checkpoint_dir, 'epoch_*'))
        ckpts = [i for i in ckpts if os.path.exists(
            pj(self.config.checkpoint_dir, 'optimizers', f"optimizer.{i.split('epoch_')[-1]}.pt"))]
        flag = False
        if len(ckpts):
            epochs = sorted([int(i.split('epoch_')[-1]) for i in ckpts], reverse=True)
            for epoch in epochs:
                try:
                    path = pj(self.config.checkpoint_dir, f"epoch_{epoch}")
                    logging.info(f'load checkpoint from {path}')
                    self.model = TransformersQG(
                        model=path, max_length=self.config.max_length, max_length_output=self.config.max_length_output,
                        label_smoothing=self.config.label_smoothing, add_prefix=add_prefix,
                        drop_overflow_error_text=True
                    )
                    self.optimizer = self.setup_optimizer(epoch)
                    self.current_epoch = epoch
                    assert self.current_epoch <= self.config.epoch, 'model training is done'
                    flag = True
                except Exception:
                    logging.exception(f'error at loading checkpoint {ckpts}')
                if flag:
                    break
        if not flag:
            logging.info(f'initialize checkpoint with {self.config.model}')
            self.model = TransformersQG(
                model=self.config.model, max_length=self.config.max_length,
                max_length_output=self.config.max_length_output, add_prefix=add_prefix,
                drop_overflow_error_text=True)
            self.optimizer = self.setup_optimizer()
            self.current_epoch = 0

        # GPU mixture precision
        self.scaler = torch.cuda.amp.GradScaler(enabled=self.config.fp16)

        # cached data folder
        input_types = to_list(self.config.input_types, sorting=False)
        output_types = to_list(self.config.output_types, sorting=False)
        assert len(input_types) == len(output_types)
        if prefix_types is None:
            prefix_types = [None] * len(input_types)
        else:
            prefix_types = to_list(self.config.prefix_types, sorting=False)
        prefix = pj(
            DEFAULT_CACHE_DIR,
            "encoded_feature"
            f"{self.config.dataset_path}{'.' + self.config.dataset_name if self.config.dataset_name != 'default' else ''}",
            f"{self.config.model}.{self.config.max_length}.{self.config.max_length_output}"
        )

        self.data_cache_paths = [[(i, o, p), f'{prefix}.{i}.{o}.train.{p}.pkl']
                                 for i, o, p in zip(input_types, output_types, prefix_types)]

    def setup_optimizer(self, epoch: int = None):
        # optimizer
        optimizer = torch.optim.AdamW(self.model.model.parameters(), lr=self.config.lr)
        if epoch is not None:
            # load from existing config
            path = pj(self.config.checkpoint_dir, "optimizers", f'optimizer.{epoch}.pt')
            logging.info(f'load optimizer from {path}')
            device = 'cpu' if OPTIMIZER_ON_CPU else self.model.device
            logging.info(f'optimizer is loading on {device}')
            optimizer_stat = torch.load(path, map_location=torch.device(device))
            optimizer.load_state_dict(optimizer_stat['optimizer_state_dict'])
            del optimizer_stat
        return optimizer

    def save(self, current_epoch):
        # save model
        save_dir = pj(self.config.checkpoint_dir, f'epoch_{current_epoch + 1}')
        os.makedirs(save_dir, exist_ok=True)
        logging.info('saving model related files')
        self.model.save(save_dir)
        with open(pj(save_dir, 'trainer_config.json'), 'w') as f:
            tmp = self.config.config.copy()
            tmp['epoch'] = current_epoch + 1
            json.dump(obj=tmp, fp=f)

        # save optimizer
        save_dir_opt = pj(self.config.checkpoint_dir, 'optimizers', f'optimizer.{current_epoch + 1}.pt')
        os.makedirs(os.path.dirname(save_dir_opt), exist_ok=True)
        logging.info('saving optimizer')
        torch.save({'optimizer_state_dict': self.optimizer.state_dict()}, save_dir_opt)

        logging.info('remove old optimizer files')
        path = pj(self.config.checkpoint_dir, 'optimizers', f'optimizer.{current_epoch}.pt')
        if os.path.exists(path):
            os.remove(path)

    def train(self, epoch_save: None or int = 1, interval: int = 25, epoch_partial: int = None):
        """ Train model.

        @param epoch_save: Save the model every this epoch.
        @param interval:
        @param epoch_partial:
        """
        self.model.train()

        if self.current_epoch == self.config.epoch:
            logging.info('training is completed')
            return None

        logging.info('dataset preprocessing')
        encode_list = []
        for (i, o, p), cache_path in self.data_cache_paths:
            text_input, text_output = get_dataset(
                self.config.dataset_path, self.config.dataset_name, split='train', input_type=i, output_type=o,
                use_auth_token=self.use_auth_token)
            encode_list += self.model.text_to_encode(text_input, text_output, prefix_type=p, cache_path=cache_path)
        loader = self.model.get_data_loader(encode_list, batch_size=self.config.batch, shuffle=True, drop_last=True)

        logging.info('start model training')
        global_step = 0
        saved_checkpoints = []
        with torch.cuda.amp.autocast(enabled=self.config.fp16):
            for e in range(self.current_epoch, self.config.epoch):  # loop over the epoch

                # # free the cache
                # gc.collect()
                # torch.cuda.empty_cache()

                mean_loss, global_step = self.train_single_epoch(loader, global_step, interval)
                logging.info(f"[epoch {e}/{self.config.epoch}] average loss: {round(mean_loss, 3)}, "
                             f"lr: {self.optimizer.param_groups[0]['lr']}")
                if epoch_save is not None and (e + 1) % epoch_save == 0 and (e + 1) != 0:
                    self.save(e)
                    saved_checkpoints.append(e)
                if epoch_partial is not None and (e + 1) == epoch_partial:
                    break
        if e not in saved_checkpoints:
            self.save(e)
        logging.info(f'complete training: model ckpt was saved at {self.config.checkpoint_dir}')

    def train_single_epoch(self, data_loader, global_step: int, interval):
        total_loss = []
        self.optimizer.zero_grad()
        for n, encode in enumerate(data_loader):

            # gc.collect()
            # torch.cuda.empty_cache()

            loss = self.model.encode_to_loss(encode)
            self.scaler.scale(loss).backward()
            total_loss.append(loss.cpu().item())
            if (n + 1) % self.config.gradient_accumulation_steps != 0:
                continue

            global_step += 1
            _total_loss = total_loss[-self.config.gradient_accumulation_steps:]
            inst_loss = sum(_total_loss)/len(_total_loss)

            # optimizer update
            self.scaler.step(self.optimizer)
            self.scaler.update()
            self.optimizer.zero_grad()
            if global_step % interval == 0:
                logging.info(f"\t * (global step {global_step}: loss: {inst_loss}, "
                             f"lr: {self.optimizer.param_groups[0]['lr']}")
        self.optimizer.zero_grad()
        return sum(total_loss)/len(total_loss), global_step



""" SQuAD QG evaluation (sentence/answer level) """
import logging
import argparse
import json

from lmqg import evaluate


logging.basicConfig(format='%(asctime)s %(levelname)-8s %(message)s', level=logging.INFO, datefmt='%Y-%m-%d %H:%M:%S')


def get_options():
    parser = argparse.ArgumentParser(description='QG evaluation on SQuAD.')
    parser.add_argument('-m', '--model-checkpoint', default=None, type=str)
    parser.add_argument('--batch-size', default=16, type=int)
    parser.add_argument('--n-beams', default=4, type=int)
    parser.add_argument('-e', '--export-dir', required=True, type=str)
    parser.add_argument('--hyp-test', default=None, type=str)
    parser.add_argument('--hyp-dev', default=None, type=str)
    parser.add_argument('--max-length', default=512, type=int, help='')
    parser.add_argument('--max-length-output', default=64, type=int, help='')
    parser.add_argument('-d', '--dataset-path', help='huggingface datasets alias', default='lmqg/qg_squad', type=str)
    parser.add_argument('--dataset-name', help='huggingface datasets name', default='default', type=str)
    parser.add_argument('-i', '--input-type', help='', default='paragraph_answer', type=str)
    parser.add_argument('-o', '--output-type', help='', default='question', type=str)
    parser.add_argument('-l', '--language', help='', default='en', type=str)
    parser.add_argument('--test-split', help='the name of test split', default='test', type=str)
    parser.add_argument('--validation-split', help='the name of validation split', default='validation', type=str)
    parser.add_argument('--overwrite', help='', action='store_true')
    parser.add_argument('--bleu-only', help='', action='store_true')
    parser.add_argument('--use-auth-token', help='', action='store_true')
    parser.add_argument('--prediction-aggregation', default=None, type=str)
    parser.add_argument('--prediction-level', default=None, type=str)

    return parser.parse_args()


def main():
    opt = get_options()
    prediction_aggregation = 'first' if opt.prediction_aggregation is None else opt.prediction_aggregation
    prediction_level = 'sentence'  # qg
    if opt.input_type == 'paragraph' and opt.output_type == 'questions_answers':  # e2e qag
        prediction_level = 'answer'
    elif opt.output_type == 'answer':  # answer extraction, question answering
        prediction_level = 'answer'
    prediction_level = opt.prediction_level if opt.prediction_level is not None else prediction_level
    assert opt.model_checkpoint or opt.hyp_test or opt.hyp_dev
    assert prediction_aggregation in ['first', 'last', 'long', 'short', 'middle']
    metric = evaluate(
        export_dir=opt.export_dir,
        batch_size=opt.batch_size,
        n_beams=opt.n_beams,
        hypothesis_file_dev=opt.hyp_dev,
        hypothesis_file_test=opt.hyp_test,
        model=opt.model_checkpoint,
        max_length=opt.max_length,
        max_length_output=opt.max_length_output,
        dataset_path=opt.dataset_path,
        dataset_name=opt.dataset_name,
        input_type=opt.input_type,
        output_type=opt.output_type,
        prediction_aggregation=prediction_aggregation,
        prediction_level=prediction_level,
        overwrite=opt.overwrite,
        language=opt.language,
        bleu_only=opt.bleu_only,
        use_auth_token=opt.use_auth_token,
    )
    print(json.dumps(metric, indent=4, sort_keys=True))




""" Language model for Huggingface inference API. """
import requests
import json
import logging
import re
from typing import List

import transformers
from .exceptions import APIError, HighlightNotFoundError, AnswerNotFoundError
from .spacy_module import SpacyPipeline
from .language_model import clean, TASK_PREFIX, ADDITIONAL_SP_TOKENS


def call_api(input_text,
             api_token,
             model,
             max_length: int = 64,
             do_sample: bool = True,
             num_beams: int = 1,
             top_p: float = 0.9,
             use_gpu: bool = False):
    data = json.dumps({
        "inputs": input_text,
        "parameters": {"do_sample": do_sample, "num_beams": num_beams, "max_length": max_length, "top_p": top_p},
        "options": {"wait_for_model": True, "use_gpu": use_gpu}
    })
    response = requests.request("POST", f"https://api-inference.huggingface.co/models/{model}",
                                headers={"Authorization": f"Bearer {api_token}"}, data=data)
    data = response.json()
    if type(data) is dict and 'error' in data.keys():
        raise APIError(data)
    return data


def highlight_sentence(input_text: str, highlight_span: str, prefix: str = None):
    position = input_text.find(highlight_span)
    if position == -1:
        HighlightNotFoundError(highlight_span, input_text)
    input_text = f"{input_text[:position]}{ADDITIONAL_SP_TOKENS['hl']} " \
                 f"{highlight_span} {ADDITIONAL_SP_TOKENS['hl']}{input_text[position + len(highlight_span):]}"
    if prefix is not None:
        input_text = f"{prefix}: {input_text}"
    return input_text


def generate_qa(
        api_token: str,
        input_text: str,
        model_qg: str,
        spacy: SpacyPipeline = None,
        model_ae: str = None,
        input_answer: List or str = None,
        max_length: int = 256,
        do_sample: bool = True,
        num_beams: int = 1,
        top_p: float = 0.9,
        use_gpu: bool = False,
        is_qag: bool = None,
        add_prefix_qg: bool = None,
        add_prefix_answer: bool = None,
        splitting_symbol: str = '|',
        question_prefix: str = "question: ",
        answer_prefix: str = ", answer: "):

    logging.info('initialize TransformersQGInferenceAPI')
    is_qag = model_qg.endswith('qag') if is_qag is None else is_qag
    add_prefix_qg = transformers.AutoConfig.from_pretrained(model_qg).add_prefix if add_prefix_qg is None else add_prefix_qg
    if is_qag:
        assert input_answer is None
        input_text = f"{TASK_PREFIX['qag'] if add_prefix_qg else None}: {input_text}"
        input_text = f"{TASK_PREFIX['qag']}: {input_text}" if add_prefix_qg else input_text
        output = call_api(
            input_text=input_text,
            api_token=api_token,
            model=model_qg,
            max_length=max_length,
            do_sample=do_sample,
            num_beams=num_beams,
            use_gpu=use_gpu,
            top_p=top_p)

        qa = []
        for raw_string in output[0]['generated_text'].split(splitting_symbol):
            if len(raw_string.split(answer_prefix)) != 2 or question_prefix not in raw_string:
                logging.info(f"invalid prediction: {raw_string}")
            else:
                q, a = raw_string.split(answer_prefix)
                a = re.sub(r'\A\s+', '', a)
                a = re.sub(r'\s+\Z', '', a)
                q = q.replace(question_prefix, "")
                q = re.sub(r'\A\s+', '', q)
                q = re.sub(r'\s+\Z', '', q)
                qa.append({'question': q, 'answer': a})
    else:

        if input_answer is None or len(input_answer) == 0:
            assert spacy is not None
            logging.info(f"answer extraction: {spacy.algorithm if model_ae is None else model_ae}")
            if model_ae is None:  # keyword extraction
                input_answer = spacy.keyword(input_text)
            else:
                add_prefix_ae = transformers.AutoConfig.from_pretrained(model_ae).add_prefix if add_prefix_answer is None else add_prefix_answer
                batch = [highlight_sentence(input_text, i, TASK_PREFIX['ae'] if add_prefix_ae else None) for i in spacy.sentence(input_text)]
                output = call_api(
                    input_text=batch,
                    api_token=api_token,
                    model=model_ae,
                    max_length=max_length,
                    do_sample=do_sample,
                    num_beams=num_beams,
                    use_gpu=use_gpu,
                    top_p=top_p)
                input_answer = list(filter(None, [clean(i['generated_text']) for i in output]))  # remove None
                input_answer = list(filter(lambda x: x in input_text, input_answer))  # remove answers not in input_text
                if len(input_answer) == 0:
                    raise AnswerNotFoundError(input_text)

        logging.info("generate question")
        input_answer = [input_answer] if type(input_answer) is str else input_answer
        batch = [highlight_sentence(input_text, i, TASK_PREFIX['qg'] if add_prefix_qg else None) for i in input_answer]
        output = call_api(
            input_text=batch,
            api_token=api_token,
            model=model_qg,
            max_length=max_length,
            do_sample=do_sample,
            num_beams=num_beams,
            top_p=top_p,
            use_gpu=use_gpu)
        question = [i['generated_text'] for i in output]
        assert len(question) == len(input_answer), f"{question} != {input_answer}"
        qa = [{'question': q, 'answer': a} for q, a in zip(question, input_answer)]
        logging.info(f"complete process: {len(output)} qa pairs generated")
    return qa





#!/usr/bin/env python
# coding=utf-8
# Copyright 2020 The HuggingFace Team All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# https://github.com/huggingface/datasets/issues/1942
"""
Fine-tuning the library models for question answering using a slightly adapted version of the 🤗 Trainer.
"""
# You can also adapt this script on your own question answering task. Pointers for this are left as comments.

import logging
import os
import urllib
import shutil
import json
import multiprocessing
import torch
from typing import Dict
from os.path import join as pj
from datasets import load_dataset, load_metric
from transformers import (
    AutoConfig,
    AutoModelForQuestionAnswering,
    AutoTokenizer,
    EvalPrediction,
    PreTrainedTokenizerFast,
    TrainingArguments,
    default_data_collator,
    set_seed,
)
from transformers.utils import check_min_version
from transformers.utils.versions import require_version
import ray
from ray import tune
from huggingface_hub import create_repo
import torch.distributed as dist

from .trainer_qa import QuestionAnsweringTrainer
from .utils_qa import postprocess_qa_predictions


os.environ["WANDB_DISABLED"] = "true"

# Will error if the minimal version of Transformers is not installed. Remove at your own risks.
check_min_version("4.18.0.dev0")
require_version("datasets>=1.8.0", "To fix: pip install -r examples/pytorch/question-answering/requirements.txt")


def internet_connection(host='http://google.com'):
    try:
        urllib.request.urlopen(host)
        return True
    except:
        return False


def run_qa_evaluation(dataset: str,
                      dataset_name: str = None,
                      dataset_files: Dict = None,
                      language_model: str = "distilbert-base-uncased",
                      eval_step: int = 50,
                      random_seed: int = 42,
                      split_train: str = 'train',
                      split_validation: str = 'validation',
                      split_test: str = 'test',
                      parallel: bool = False,
                      n_trials: int = 10,
                      ray_result_dir: str = 'ray_results',
                      output_dir: str = 'qa_eval_output',
                      overwrite: bool = False,
                      skip_training: bool = False,
                      skip_test: bool = False,
                      doc_stride: int = 128,
                      max_seq_length: int = 384,
                      max_answer_length: int = 30,
                      question_column_name: str = "question",
                      context_column_name: str = "context",
                      answer_column_name: str = "answers",
                      hf_model_alias_to_push: str = None,
                      hf_organization_to_push: str = None,
                      hf_use_auth_token: bool = False,
                      down_sample_size_train: int = None,
                      num_cpus: int = 4,
                      down_sample_size_validation: int = None):
    ray.init(ignore_reinit_error=True, num_cpus=num_cpus)
    best_hyperparameters_path = pj(output_dir, 'best_hyperparameters.json')
    best_model_path = pj(output_dir, 'best_model')
    summary_file = pj(output_dir, 'test_result.json')
    os.makedirs(output_dir, exist_ok=True)
    # Set seed before initializing model.
    set_seed(random_seed)
    local_file_only = not internet_connection()
    if dataset_files is None:
        raw_datasets = load_dataset(dataset) if dataset_name is None else load_dataset(dataset, dataset_name)
    else:
        raw_datasets = load_dataset(dataset, data_files=dataset_files)

    # Load pretrained model and tokenizer
    config = AutoConfig.from_pretrained(language_model, local_files_only=local_file_only)
    tokenizer = AutoTokenizer.from_pretrained(language_model, use_fast=True, local_files_only=local_file_only)
    model = AutoModelForQuestionAnswering.from_pretrained(language_model, config=config, local_files_only=local_file_only)

    # Tokenizer check: this script requires a fast tokenizer.
    if not isinstance(tokenizer, PreTrainedTokenizerFast):
        raise ValueError(
            "This example script only works for models that have a fast tokenizer. Checkout the big table of models "
            "at https://huggingface.co/transformers/index.html#supported-frameworks to find the model types that meet this "
            "requirement"
        )

    if max_seq_length > tokenizer.model_max_length:
        logging.warning(
            f"The max_seq_length passed ({max_seq_length}) is larger than the maximum length for the"
            f"model ({tokenizer.model_max_length}). Using max_seq_length={tokenizer.model_max_length}."
        )
    max_seq_length = min(max_seq_length, tokenizer.model_max_length)
    # Padding side determines if we do (question|context) or (context|question).
    pad_on_right = tokenizer.padding_side == "right"

    # Training preprocessing
    def prepare_train_features(examples):
        # Some of the questions have lots of whitespace on the left, which is not useful and will make the
        # truncation of the context fail (the tokenized question will take a lots of space). So we remove that
        # left whitespace
        examples[question_column_name] = [q.lstrip() for q in examples[question_column_name]]

        # Tokenize our examples with truncation and maybe padding, but keep the overflows using a stride. This results
        # in one example possible giving several features when a context is long, each of those features having a
        # context that overlaps a bit the context of the previous feature.
        tokenized_examples = tokenizer(
            examples[question_column_name if pad_on_right else context_column_name],
            examples[context_column_name if pad_on_right else question_column_name],
            truncation="only_second" if pad_on_right else "only_first",
            max_length=max_seq_length,
            stride=doc_stride,
            return_overflowing_tokens=True,
            return_offsets_mapping=True,
            padding="max_length"
        )

        # Since one example might give us several features if it has a long context, we need a map from a feature to
        # its corresponding example. This key gives us just that.
        sample_mapping = tokenized_examples.pop("overflow_to_sample_mapping")
        # The offset mappings will give us a map from token to character position in the original context. This will
        # help us compute the start_positions and end_positions.
        offset_mapping = tokenized_examples.pop("offset_mapping")

        # Let's label those examples!
        tokenized_examples["start_positions"] = []
        tokenized_examples["end_positions"] = []

        for i, offsets in enumerate(offset_mapping):
            # We will label impossible answers with the index of the CLS token.
            input_ids = tokenized_examples["input_ids"][i]
            cls_index = input_ids.index(tokenizer.cls_token_id)

            # Grab the sequence corresponding to that example (to know what is the context and what is the question).
            sequence_ids = tokenized_examples.sequence_ids(i)

            # One example can give several spans, this is the index of the example containing this span of text.
            sample_index = sample_mapping[i]
            answers = examples[answer_column_name][sample_index]
            # If no answers are given, set the cls_index as answer.
            if len(answers["answer_start"]) == 0:
                tokenized_examples["start_positions"].append(cls_index)
                tokenized_examples["end_positions"].append(cls_index)
            else:
                # Start/end character index of the answer in the text.
                start_char = answers["answer_start"][0]
                end_char = start_char + len(answers["text"][0])

                # Start token index of the current span in the text.
                token_start_index = 0
                while sequence_ids[token_start_index] != (1 if pad_on_right else 0):
                    token_start_index += 1

                # End token index of the current span in the text.
                token_end_index = len(input_ids) - 1
                while sequence_ids[token_end_index] != (1 if pad_on_right else 0):
                    token_end_index -= 1

                # Detect if the answer is out of the span (in which case this feature is labeled with the CLS index).
                if answers["text"][0] == "":
                    pass
                elif not (offsets[token_start_index][0] <= start_char and offsets[token_end_index][1] >= end_char):
                    logging.warning(f"answer not found:\n \t - answer: {answers}\n \t - context: "
                                    f"{tokenizer.decode(tokenized_examples['input_ids'][i])}")
                    tokenized_examples["start_positions"].append(cls_index)
                    tokenized_examples["end_positions"].append(cls_index)
                else:
                    try:
                        # Otherwise move the token_start_index and token_end_index to the two ends of the answer.
                        # Note: we could go after the last offset if the answer is the last word (edge case).
                        while token_start_index < len(offsets) and offsets[token_start_index][0] <= start_char:
                            token_start_index += 1
                        tokenized_examples["start_positions"].append(token_start_index - 1)

                        while offsets[token_end_index][1] >= end_char:
                            token_end_index -= 1
                        tokenized_examples["end_positions"].append(token_end_index + 1)
                        answer_found = tokenizer.decode(
                            tokenized_examples["input_ids"][i][
                            tokenized_examples["start_positions"][i]: 1 + tokenized_examples["end_positions"][i]]
                        )
                        if answers['text'][0].lower().replace(" ", "") != answer_found.lower().replace(" ", ""):
                            logging.debug(f"answer not matched:\n \t - reference: {answers['text'][0].lower()}\n \t "
                                          f"- found: {answer_found.lower()}")
                    except Exception:
                        logging.warning(f"answer not found:\n \t - answer: {answers}\n \t - context: "
                                        f"{tokenizer.decode(tokenized_examples['input_ids'][i])}")
                        tokenized_examples["start_positions"].append(cls_index)
                        tokenized_examples["end_positions"].append(cls_index)

        return tokenized_examples

    # Validation preprocessing
    def prepare_validation_features(examples):
        # Some of the questions have lots of whitespace on the left, which is not useful and will make the
        # truncation of the context fail (the tokenized question will take a lots of space). So we remove that
        # left whitespace
        examples[question_column_name] = [q.lstrip() for q in examples[question_column_name]]

        # Tokenize our examples with truncation and maybe padding, but keep the overflows using a stride. This results
        # in one example possible giving several features when a context is long, each of those features having a
        # context that overlaps a bit the context of the previous feature.
        tokenized_examples = tokenizer(
            examples[question_column_name if pad_on_right else context_column_name],
            examples[context_column_name if pad_on_right else question_column_name],
            truncation="only_second" if pad_on_right else "only_first",
            max_length=max_seq_length,
            stride=doc_stride,
            return_overflowing_tokens=True,
            return_offsets_mapping=True,
            padding="max_length"
        )

        # Since one example might give us several features if it has a long context, we need a map from a feature to
        # its corresponding example. This key gives us just that.
        sample_mapping = tokenized_examples.pop("overflow_to_sample_mapping")

        # For evaluation, we will need to convert our predictions to substrings of the context, so we keep the
        # corresponding example_id and we will store the offset mappings.
        tokenized_examples["example_id"] = []

        for i in range(len(tokenized_examples["input_ids"])):
            # Grab the sequence corresponding to that example (to know what is the context and what is the question).
            sequence_ids = tokenized_examples.sequence_ids(i)
            context_index = 1 if pad_on_right else 0

            # One example can give several spans, this is the index of the example containing this span of text.
            sample_index = sample_mapping[i]
            tokenized_examples["example_id"].append(examples["id"][sample_index])

            # Set to None the offset_mapping that are not part of the context so it's easy to determine if a token
            # position is part of the context or not.
            tokenized_examples["offset_mapping"][i] = [
                (o if sequence_ids[k] == context_index else None)
                for k, o in enumerate(tokenized_examples["offset_mapping"][i])
            ]

        return tokenized_examples

    # Create train feature from dataset
    # train_example = raw_datasets[split_train].filter(lambda x: len(x['question']) < max_seq_length)
    train_example = raw_datasets[split_train]
    train_dataset = train_example.map(
        prepare_train_features, batched=True, num_proc=None,
        remove_columns=train_example.column_names, desc="Running tokenizer on train dataset"
    )
    if down_sample_size_train is not None and down_sample_size_train < len(train_example):
        train_example_search = train_example.shuffle(random_seed)
        train_example_search = train_example_search.select(list(range(down_sample_size_train)))
        train_dataset_search = train_example_search.map(
            prepare_train_features, batched=True, num_proc=None,
            remove_columns=train_example.column_names, desc="Running tokenizer on train dataset"
        )
    else:
        train_dataset_search = train_dataset

    # Validation Feature Creation
    # validation_example = raw_datasets[split_validation].filter(lambda x: len(x['question']) < max_seq_length)
    validation_example = raw_datasets[split_validation]
    if down_sample_size_validation is not None and down_sample_size_validation < len(validation_example):
        validation_example = validation_example.shuffle(random_seed)
        validation_example = validation_example.select(list(range(down_sample_size_validation)))
    validation_dataset = validation_example.map(
        prepare_validation_features, batched=True, num_proc=None,
        remove_columns=validation_example.column_names, desc="Running tokenizer on validation dataset",
    )

    # Predict Feature Creation
    if split_test in raw_datasets:
        test_example = raw_datasets[split_test]
        test_dataset = test_example.map(
            prepare_validation_features, batched=True, num_proc=None,
            remove_columns=test_example.column_names, desc="Running tokenizer on prediction dataset"
        )
    else:
        test_example = test_dataset = None

    # Post-processing:
    def post_processing_function(examples, features, predictions, stage="eval"):
        # Post-processing: we match the start logits and end logits to answers in the original context.
        predictions = postprocess_qa_predictions(
            examples=examples,
            features=features,
            predictions=predictions,
            version_2_with_negative=False,
            n_best_size=20,
            max_answer_length=max_answer_length,
            null_score_diff_threshold=0.0,
            output_dir=output_dir,
            prefix=stage
        )

        # Format the result to the format the metric expects.
        formatted_predictions = [{"id": k, "prediction_text": _v} for k, _v in predictions.items()]
        references = [{"id": ex["id"], "answers": ex[answer_column_name]} for ex in examples]
        return EvalPrediction(predictions=formatted_predictions, label_ids=references)

    if dist.is_initialized():
        metric = load_metric("squad", num_process=dist.get_world_size(), process_id=dist.get_rank())
    else:
        metric = load_metric("squad")

    def compute_metrics(p: EvalPrediction):
        return metric.compute(predictions=p.predictions, references=p.label_ids)

    if not skip_training and (not os.path.exists(best_model_path) or overwrite):
        # Initialize our Trainer
        training_args = TrainingArguments(
            report_to=None,
            output_dir=output_dir,
            eval_steps=eval_step,
            seed=random_seed,
            evaluation_strategy="steps"
        )
        trainer = QuestionAnsweringTrainer(
            model=model,
            args=training_args,
            train_dataset=train_dataset_search,
            eval_dataset=validation_dataset,
            eval_examples=validation_example,
            tokenizer=tokenizer,
            data_collator=default_data_collator,
            post_process_function=post_processing_function,
            compute_metrics=compute_metrics,
            model_init=lambda x: AutoModelForQuestionAnswering.from_pretrained(language_model, return_dict=True, local_files_only=local_file_only)
        )
        if not os.path.exists(best_hyperparameters_path) or overwrite:

            logging.info("*** Finetuning ***")
            if parallel:
                best_run = trainer.hyperparameter_search(
                    hp_space=lambda x: {
                        "learning_rate": tune.loguniform(1e-6, 1e-4),
                        "num_train_epochs": tune.choice(list(range(1, 6))),
                        "per_device_train_batch_size": tune.choice([4, 8, 16, 32, 64]),
                    },
                    local_dir=ray_result_dir, direction="maximize", backend="ray", n_trials=n_trials,
                    resources_per_trial={'cpu': multiprocessing.cpu_count(), "gpu": torch.cuda.device_count()}
                )
            else:
                best_run = trainer.hyperparameter_search(
                    hp_space=lambda x: {
                        "learning_rate": tune.loguniform(1e-6, 1e-4),
                        "num_train_epochs": tune.choice(list(range(1, 6))),
                        "per_device_train_batch_size": tune.choice([4, 8, 16, 32, 64]),
                    },
                    local_dir=ray_result_dir, direction="maximize", backend="ray", n_trials=n_trials,
                    resources_per_trial={'cpu': 1, "gpu": torch.cuda.device_count()}
                )
            best_run_hyperparameters = best_run.hyperparameters
            logging.info("best config")
            logging.info(json.dumps(best_run_hyperparameters, indent=4))
            with open(best_hyperparameters_path, 'w') as f:
                json.dump(best_run_hyperparameters, f)
        with open(best_hyperparameters_path) as f:
            best_run_hyperparameters = json.load(f)
        for n, v in best_run_hyperparameters.items():
            setattr(trainer.args, n, v)
        # update training dataset to full dataset without down sampling
        setattr(trainer, 'train_dataset', train_dataset)
        assert trainer.train_dataset == train_dataset
        # train
        trainer.train()
        trainer.save_model(best_model_path)
        tokenizer.save_pretrained(best_model_path)

    # Evaluation
    if not skip_test and test_example is not None and test_dataset is not None:
        if not os.path.exists(summary_file) or overwrite:
            logging.info("*** Evaluate ***")
            trainer = QuestionAnsweringTrainer(
                model=AutoModelForQuestionAnswering.from_pretrained(best_model_path, local_files_only=local_file_only),
                args=TrainingArguments(report_to=None, output_dir=output_dir, seed=random_seed, evaluation_strategy="no"),
                eval_dataset=test_dataset,
                eval_examples=test_example,
                tokenizer=tokenizer,
                data_collator=default_data_collator,
                post_process_function=post_processing_function,
                compute_metrics=compute_metrics,
            )
            metric = trainer.evaluate()
            result = {k: v for k, v in metric.items()}
            logging.info(json.dumps(result, indent=4))
            with open(summary_file, 'w') as f:
                json.dump(result, f)

    if hf_organization_to_push is not None and hf_model_alias_to_push is not None:
        url = create_repo(hf_model_alias_to_push, organization=hf_organization_to_push, exist_ok=True)
        args = {"use_auth_token": hf_use_auth_token, "repo_url": url, "organization": hf_organization_to_push}
        model.push_to_hub(hf_model_alias_to_push, **args)
        tokenizer.push_to_hub(hf_model_alias_to_push, **args)
        readme = "This model is automatically pushed by [lmqg](https://github.com/asahi417/lm-question-generation) library."
        with open(f"{hf_model_alias_to_push}/README.md", "w") as f:
            f.write(readme)
        if os.path.exists(best_hyperparameters_path):
            shutil.copy2(best_hyperparameters_path, pj(hf_model_alias_to_push, 'best_run_hyperparameters.json'))
        if os.path.exists(summary_file):
            shutil.copy2(summary_file, pj(hf_model_alias_to_push, 'metric.json'))
        os.system(
            f"cd {hf_model_alias_to_push} && git lfs install && git add . && git commit -m 'model update' && git push && cd ../")



import argparse
import logging
from lmqg.qa_evaluation_tool import generate_qa_pairs, run_qa_evaluation

logging.basicConfig(format='%(asctime)s %(levelname)-8s %(message)s', level=logging.INFO, datefmt='%Y-%m-%d %H:%M:%S')


def main_qa_model_training():
    parser = argparse.ArgumentParser(description='QA model training.')
    parser.add_argument('-m', '--model', default='distilbert-base-uncased', type=str)
    parser.add_argument('-d', '--dataset', default='lmqg/qa_squadshifts', type=str)
    parser.add_argument('-n', '--dataset-name', default=None, type=str)
    parser.add_argument('--dataset-train', default=None, type=str)
    parser.add_argument('--dataset-validation', default=None, type=str)
    parser.add_argument('--dataset-test', default=None, type=str)
    parser.add_argument('--eval-step', default=50, type=int)
    parser.add_argument('--random-seed', default=42, type=int)
    parser.add_argument('--split-train', default='train', type=str)
    parser.add_argument('--split-validation', default='validation', type=str)
    parser.add_argument('--split-test', default='test', type=str)
    parser.add_argument('--parallel', action='store_true')
    parser.add_argument('--n-trials', default=10, type=int)
    parser.add_argument('--max-seq-length', default=384, type=int)
    parser.add_argument('--ray-result-dir', default='ray_results', type=str)
    parser.add_argument('--output-dir', default='qa_eval_output', type=str)
    parser.add_argument('--overwrite', action='store_true')
    parser.add_argument('--skip-training', action='store_true')
    parser.add_argument('--down-sample-size-train', default=None, type=int)
    parser.add_argument('--down-sample-size-validation', default=None, type=int)
    parser.add_argument('--num-cpus', default=4, type=int)

    opt = parser.parse_args()
    if opt.dataset_train is not None and opt.dataset_validation is not None and opt.dataset_test is not None:
        dataset_files = {
            'train': opt.dataset_train,
            'validation': opt.dataset_validation,
            'test': opt.dataset_test
        }
    else:
        dataset_files = None

    run_qa_evaluation(
        dataset=opt.dataset,
        dataset_name=opt.dataset_name,
        dataset_files=dataset_files,
        language_model=opt.model,
        eval_step=opt.eval_step,
        random_seed=opt.random_seed,
        max_seq_length=opt.max_seq_length,
        split_train=opt.split_train,
        split_validation=opt.split_validation,
        split_test=opt.split_test,
        parallel=opt.parallel,
        n_trials=opt.n_trials,
        ray_result_dir=opt.ray_result_dir,
        output_dir=opt.output_dir,
        overwrite=opt.overwrite,
        skip_training=opt.skip_training,
        down_sample_size_train=opt.down_sample_size_train,
        down_sample_size_validation=opt.down_sample_size_validation,
        num_cpus=opt.num_cpus
    )


def main_generate_qa_pair():
    parser = argparse.ArgumentParser(description='Generate QA pseudo dataset.')
    parser.add_argument('-m', '--model-qg', default='lmqg/t5-small-squad-qg', type=str)
    parser.add_argument('--model-ae', default=None, type=str)
    parser.add_argument('-l', '--language', default='en', type=str)
    parser.add_argument('-d', '--anchor-data', default='lmqg/qa_squadshifts', type=str)
    parser.add_argument('-n', '--anchor-data-name', default='new_wiki', type=str)
    parser.add_argument('--use-reference-answer', action='store_true')
    parser.add_argument('--overwrite', action='store_true')
    parser.add_argument('-b', '--batch-size', default=256, type=int)
    parser.add_argument('-e', '--export-dir', default='tmp_output', type=str)
    parser.add_argument('--max-length', default=512, type=int, help='')
    parser.add_argument('--max-length-output', default=64, type=int, help='')
    opt = parser.parse_args()
    generate_qa_pairs(
        model_qg=opt.model_qg,
        model_ae=opt.model_ae,
        language=opt.language,
        anchor_data=opt.anchor_data,
        anchor_data_name=opt.anchor_data_name,
        use_reference_answer=opt.use_reference_answer,
        batch_size=opt.batch_size,
        export_dir=opt.export_dir,
        overwrite=opt.overwrite,
        max_length=opt.max_length,
        max_length_output=opt.max_length)
